import { PromptMemory, PromptFunctions, Tokenizer } from "promptrix";
import { SchemaBasedCommand, CommandSchema } from "../SchemaBasedCommand";
import { TaskResponse } from "../types";

const schema: CommandSchema = {
    type: "object",
    title: "confirmAnswer",
    description: "asks the user to confirm an answer before completing a task",
    properties: {
        answer: {
            type: "string",
            description: "answer to confirm"
        },
        confirmation: {
            type: "string",
            description: "confirmation question to ask the user"
        }
    },
    required: ["answer", "confirmation"],
    returns: "users confirmation or requested changes"
};

export interface ConfirmAnswerCommandInput {
    answer: string;
    confirmation: string;
}

export class ConfirmAnswerCommand extends SchemaBasedCommand<ConfirmAnswerCommandInput> {
    public constructor(title?: string, description?: string) {
        super(schema, title, description);
    }

    public execute(input: ConfirmAnswerCommandInput, memory: PromptMemory, functions: PromptFunctions, tokenizer: Tokenizer): Promise<TaskResponse> {
        return Promise.resolve({
            type: "TaskResponse",
            status: "input_needed",
            message: `${input.confirmation}\n\n${input.answer}`
        });
    }
}



from dataclasses import dataclass, asdict
from typing import Any, Dict, Optional, Union
from uuid import uuid4
from pyee import AsyncIOEventEmitter
import asyncio, copy
import json
import sys
import time
import traceback
from promptrix.promptrixTypes import Message, PromptFunctions, PromptSection, PromptMemory, Tokenizer
from promptrix.FunctionRegistry import  FunctionRegistry
from promptrix.GroupSection import GroupSection
from promptrix.GPT3Tokenizer import GPT3Tokenizer
from promptrix.VolatileMemory import VolatileMemory
from promptrix.ConversationHistory import ConversationHistory
from promptrix.TextSection import TextSection
from promptrix.AssistantMessage import AssistantMessage
from promptrix.UserMessage import UserMessage
from promptrix.Utilities import Utilities
from promptrix.TemplateSection import TemplateSection
from promptrix.Prompt import Prompt

from alphawave.AlphaWave import AlphaWave, AlphaWaveOptions
from alphawave.alphawaveTypes import PromptCompletionClient, PromptCompletionOptions, PromptResponse, Validation, PromptResponseValidator
from alphawave.JSONResponseValidator import JSONResponseValidator

from alphawave_agents.SchemaBasedCommand import SchemaBasedCommand, CommandSchema
from alphawave_agents.AgentCommandValidator import AgentCommandValidator
import promptrix.Utilities
from alphawave_agents.agentTypes import Command, AgentThought
from alphawave_agents.AgentCommandSection import AgentCommandSection

from typing import Dict, Any
from pydantic import BaseModel, Field
import traceback

class AgentCommandSchema(BaseModel):
    type: str = Field("object", description="an agent that can perform a task")
    title: str = "Agent"
    properties: Dict[str, Any] = {
        "input": {
            "type": "string",
            "description": "input for command",
        }
    }
    required: list[str] = ["input"]



@dataclass
class AgentOptions:
    client: PromptCompletionClient
    context_variable: Optional[str] = None
    prompt: Union[str, list[str], PromptSection] = None
    prompt_options: PromptCompletionOptions = None
    agent_variable: Optional[str] = None
    functions: Optional[PromptFunctions] = None
    history_variable: Optional[str] = None
    initial_thought: Optional[AgentThought] = None
    input_variable: Optional[str] = None
    logRepairs: Optional[bool] = None
    max_history_messages: Optional[int] = None
    max_repair_attempts: Optional[int] = None
    max_steps: Optional[int] = None
    memory: Optional[PromptMemory] = None
    retry_invalid_responses: Optional[bool] = None
    step_delay: Optional[int] = None
    tokenizer: Optional[Tokenizer] = None
    syntax: Optional[str] = 'JSON'

@dataclass
class ConfiguredAgentOptions:
    agent_variable: str = None
    client: PromptCompletionClient = None
    context_variable: str = None
    functions: PromptFunctions = None
    history_variable: str = None
    initial_thought: Optional[AgentThought] = None
    input_variable: str = None
    logRepairs: bool = None
    max_history_messages: int = 10
    max_repair_attempts: int = 3
    max_steps: int = 5
    memory: PromptMemory = None
    prompt: Union[str, list[str], PromptSection] = None
    prompt_options: PromptCompletionOptions = None
    retry_invalid_responses: bool = False
    step_delay: int = 5
    tokenizer: Tokenizer = None
    syntax: Optional[str] = 'JSON'

def update_dataclass(instance, source):
    for key, value in source.items():
        if key in instance and value is not None:
            instance[key] =  value

##
PromptInstructionSectionJSON = TemplateSection(\
"""You are a helpful, friendly, obedient AI.
If a question or task can be answer from known fact or reasoning respond and STOP. 
Otherwise, select a command to perform, and respond using the command format provided.
You should choose to respond directly. Remember, though, you are more than just an AI, since you have commands available.
""", 'system')

PromptOneShotJSON = [
    UserMessage("What is 35 * 64?"),
    AssistantMessage("""I will use the math command.
{"command": "math", "inputs":{"code":"35*64"}}""")]


PromptInstructionSectionTOML = TemplateSection(\
"""You are a helpful AI.
If a question or task can be completed from known fact, reasoning, or has been completed, respond immediately. 
Otherwise, develop a concise plan for completing the task. Base your plan on known fact, reasoning, prior results, and the available commands.
Select the first step of that plan as the next command to perform, and respond with the command format provided to invoke the command.
""",'system')

PromptOneShotTOML = [
    UserMessage("What is 35 * 64?"),
    AssistantMessage("""I will use the math command.
[RESPONSE]
command = "math"
inputs.code = "35*64"
[STOP]
""")]


@dataclass
class AgentCommandInput:
    agentId: str
    input: str


@dataclass
class AgentState:
    totalSteps: int
    context: Optional[str] = None
    child: Optional[Dict[str, str]] = None


class Agent(SchemaBasedCommand):
    def __init__(self, options: AgentOptions, title: Optional[str] = None, description: Optional[str] = None):
        super().__init__(CommandSchema, title, description)
        self._commands: Dict[str, Command] = {}
        self._options: ConfiguredAgentOptions = {
            'agent_variable': 'agent',
            'client': None,
            'context_variable': 'context',
            'functions': FunctionRegistry(),
            'history_variable': 'history',
            'initial_thought': None,
            'input_variable': 'input',
            'logRepairs': False,
            'max_history_messages': 10,
            'max_repair_attempts': 3,
            'max_steps': 5,
            'memory': VolatileMemory(),
            'prompt': None,
            'prompt_options': None,
            'retry_invalid_responses': False,
            'step_delay': 5,
            'tokenizer': GPT3Tokenizer(),
            'syntax': 'JSON'
        }
        update_dataclass(self._options, options.__dict__)
        #self._options = self._options.__dict__
        self._events: EventEmitter = AsyncIOEventEmitter()
        self.top_level_task=None

    @property
    def client(self):
        return self._options.client

    @property
    def events(self):
        return self._events

    @property
    def functions(self):
        return self._options['functions']

    @property
    def memory(self):
        return self._options['memory']

    @property
    def options(self):
        return self._options

    @property
    def tokenizer(self):
        return self._options['tokenizer']

    # Command management

    def addCommand(self, command: Command):
        if command.title in self._commands:
            raise ValueError(f"A command with the title {command.schema.title} already exists.")
        self._commands[command.title] = command
        return self

    def getCommand(self, title: str):
        return self._commands.get(title)

    def hasCommand(self, title: str):
        return title in self._commands

    def completeTask(self, input: Optional[str] = None, agentId: Optional[str] = None, executeInitialThought: bool = False):
      try:
        # Initialize the input to the next step
        stepInput = input if input is not None else self.memory.get(self.options['input_variable'])
        if self.top_level_task is None:
            self.top_level_task = stepInput

        # Dispatch to child agent if needed
        step = 0
        state = self.get_agent_state(agentId)
        # Start main task loop
        while step < self.options['max_steps']:
            # Wait for step delay to prevent gpt overrun
            if step > 0 and self._options['step_delay'] > 0:
                sys.stdout.flush()
                time.sleep(self._options['step_delay']/100) # assumes step delay is in seconds
            # Execute next step
            #print(f'***** Agent completeTask calling execute_next_step {stepInput}')
            result, ran_command = self.execute_next_step(stepInput, agentId)
            #print(f'***** Agent completeTask return from execute_next_step {ran_command}, {result}')
            if ran_command:
                # check for command in response from command
                # Create command validator
                #validator = AgentCommandValidator(self._commands, self._options['client'], self._options['prompt_options'].model,
                #                                  self._options['syntax'], memory=self.memory, history_variable=history_variable)
                # just a stub, need to finish if good idea
                return {'type':'TaskResponse', 'status':'success', 'message':result}
            else:
                return {'type':'TaskResponse', 'status':'success', 'message':result}

            step += 1
            executeInitialThought = False
        # Return too many steps
        return {
            "type": "TaskResponse",
            "status": "too_many_steps",
            "message": "The current task has taken too many steps."
        }
      except Exception as e:
        traceback.print_exc()
        pass

    def get_agent_state(self, agentId: Optional[str] = None):
        key = f'{self.options.agent_variable}-{agentId}' if agentId else self._options['agent_variable']
        state = self.memory.get(key) or {}
        if 'totalSteps' not in state:
            state['totalSteps'] = 0
        return state

    def set_agent_state(self, state: AgentState, agentId: Optional[str] = None):
        key = f"{self.options['agent_variable']}-{agentId}" if agentId else self.options['agent_variable']
        self.memory.set(key, state)

    def get_agent_history_variable(self, agentId: Optional[str] = None):
        return f"{self._options['history_variable']}-{agentId}" if agentId else self._options['history_variable']


    def execute_next_step(self, input: Optional[str] = None, agent_id: Optional[str] = None, execute_initial_thought: bool = False):
        try:
            #print(f'***** Agent execute_next_step input {input}')
            state = self.get_agent_state(agent_id)
            # Create agents prompt section
            agent_prompt = []
            if isinstance(self._options['prompt'], list):
                agent_prompt = TemplateSection('\n'.join(self._options['prompt']), 'system')
                sections = [agent_prompt]
            elif isinstance(self._options['prompt'], dict):
                agent_prompt = self._options['prompt']
                sections = [agent_prompt]
            elif isinstance(self._options['prompt'], str):
                agent_prompt = TemplateSection(self._options['prompt'], 'system')
                sections = [agent_prompt]
            
                
            # Ensure the context variable is set
            if 'context' in state:
                self.memory.set(self.options['context_variable'], state['context'])

            # Create prompt
            history_variable = self.get_agent_history_variable(agent_id)
            try:
                sections.append(AgentCommandSection(self._commands, one_shot=True, syntax=self._options['syntax']))
                
                if self._options['syntax'] == 'JSON':
                    pis = PromptInstructionSectionJSON
                    pos = PromptOneShotJSON
                else:
                    pis = PromptInstructionSectionTOML
                    pos = PromptOneShotTOML
                    
                sections.append(pis)
                #sections.extend(pos)
                prompt = Prompt([
                    GroupSection(sections, 'system'),
                    ConversationHistory(history_variable, 1.0, True)
                ])
                if input:
                    #print(f'***** Agent append Text Section {input}')
                    prompt.sections.append(TextSection(input, 'user', -1, True, '\n', 'user'))
                    # Ensure input variable is set otherwise the history will be wrong.
                    self.memory.set(self.options['input_variable'], input)
            except Exception as e:
                tracebak.print_exc()

            if execute_initial_thought and self._options['initial_thought']:
                # Just use initial thought as response
                # - This is used when agents are being called as commands.
                response = {
                    'status': 'success',
                    'message': {'role': 'assistant', 'content': self._options['initial_thought']}
                }
            else:
                # Add initial thought to history
                if state['totalSteps'] == 0 and self._options['initial_thought']:
                    history = self.memory.get(history_variable) or []
                    message = {'role': 'assistant', 'content': json.dumps(self._options['initial_thought'])}
                    history.append(message)
                    self.memory.set(history_variable, history)

                # Create command validator
                validator = AgentCommandValidator(self._commands, self._options['client'], self._options['prompt_options'].model, self._options['syntax'], memory=self.memory, history_variable=history_variable)

                # Create a wave for the prompt
                ### this is the ONLY wave call in Agent. 
                wave = AlphaWave(
                    client = self._options['client'],
                    prompt = prompt,
                    prompt_options = self._options['prompt_options'],
                    functions = self._options['functions'],
                    history_variable = history_variable,
                    input_variable = self._options['input_variable'],
                    max_history_messages = self._options['max_history_messages'],
                    max_repair_attempts = self._options['max_repair_attempts'],
                    memory = self._options['memory'],
                    tokenizer = self._options['tokenizer'],
                    logRepairs = self._options['logRepairs'],
                    validator = validator
                )

                # Complete the prompt
                max_attempts = 2 if self._options['retry_invalid_responses'] else 1
                for attempt in range(max_attempts):
                    response = wave.completePrompt()
                    
                    if response['status'] != 'invalid_response':
                        break

                # Ensure response succeeded
                # Note at this point AgentCommandValidator has approved response
                if response['status'] != 'success':
                    return {
                        'type': "TaskResponse",
                        'status': response['status'],
                        'message': response['message']
                    }, False

            # Get agents thought and execute command
            message = response['message']
            thought = message['content']

            #test if LLM wants to execute command. If no, just return!
            if type(thought) != dict or 'command' not in thought:
                #print(f'***** Agent execute_next_step returning no next command')
                return thought, False  # False means don't keep going, no command to run
            self._events.emit('newThought', thought)
            #print(f'***** Agent execute_next_step calling execute_command state {state}, thought {thought}')

            command_name = thought['command']
            command_input = str(thought['inputs'] or '')
            result = self.execute_command(state, thought)
            #print(f'command_result \n{result}\n')
            # Check for task result and error
            task_response = result if isinstance(result, dict) and result.get('type') == 'TaskResponse' else None
            if task_response:
                if task_response['status'] in ['error', 'invalid_response', 'rate_limited', 'too_many_steps', 'too_long']:
                    #print(f'***** Agent execute_next_step fail {task_response}')
                    return task_response, False

            # Update history
            return_msg = task_response['message'] if task_response else result
            history = self.memory.get(history_variable) or []
            history.append({'role': 'assistant', 'content': command_name+' '+str(input)+' result:\n'+str(return_msg)})
            self.memory.set(history_variable, history)

            # Save the agents state
            state['totalSteps'] += 1
            self.set_agent_state(state, agent_id)

            # Return result
            #return_msg = task_response if task_response else Utilities.to_string(self.tokenizer, result)
            #print(f'***** Agent execute_next_step returning {return_msg}')
            return return_msg, True
        except Exception as err:
            traceback.print_exc()
            return {
                'type': "TaskResponse",
                'status': "error",
                'message': str(err)
            }, False

    def execute_command(self, state: AgentState, thought: AgentThought):
        # a command to execute
        command_name = thought['command']
        command = self._commands.get(command_name, None) or {}
        input = thought['inputs'] or {}
        # Execute command and return result
        #print(f'***** Agent execute_command  {command_name}, {input}')
        response = command.execute(input, self.memory, self.functions, self.tokenizer)
        #print(f'***** Agent execute_command {command_name}\n{response}\n')
        return response




from dataclasses import dataclass
from typing import List, Optional, Union, Any, Dict

@dataclass
class CreateCompletionRequest:
    model: str
    prompt: Optional[Union[List[Any], List[int], List[str], str]] = None
    suffix: Optional[str] = None
    max_tokens: Optional[int] = None
    temperature: Optional[float] = None
    top_p: Optional[float] = None
    n: Optional[int] = None
    stream: Optional[bool] = None
    logprobs: Optional[int] = None
    echo: Optional[bool] = None
    stop: Optional[Union[List[str], str]] = None
    presence_penalty: Optional[float] = None
    frequency_penalty: Optional[float] = None
    best_of: Optional[int] = None
    logit_bias: Optional[Dict] = None
    user: Optional[str] = None

@dataclass
class CreateCompletionResponseChoicesInnerLogprobs:
    tokens: Optional[List[str]] = None
    token_logprobs: Optional[List[float]] = None
    top_logprobs: Optional[List[Dict]] = None
    text_offset: Optional[List[int]] = None

@dataclass
class CreateCompletionResponseChoicesInner:
    text: Optional[str] = None
    index: Optional[int] = None
    logprobs: Optional[CreateCompletionResponseChoicesInnerLogprobs] = None
    finish_reason: Optional[str] = None

@dataclass
class CreateCompletionResponseUsage:
    prompt_tokens: int
    completion_tokens: int
    total_tokens: int

@dataclass
class CreateCompletionResponse:
    id: str
    object: str
    created: int
    model: str
    choices: List[CreateCompletionResponseChoicesInner]
    usage: Optional[CreateCompletionResponseUsage] = None

@dataclass
class ChatCompletionRequestMessage:
    role: str
    content: str
    name: Optional[str] = None

@dataclass
class CreateChatCompletionRequest:
    model: str
    messages: List[ChatCompletionRequestMessage]
    temperature: Optional[float] = None
    top_p: Optional[float] = None
    n: Optional[int] = None
    stream: Optional[bool] = None
    stop: Optional[Union[List[str], str]] = None
    max_tokens: Optional[int] = None
    presence_penalty: Optional[float] = None
    frequency_penalty: Optional[float] = None
    logit_bias: Optional[Dict] = None
    user: Optional[str] = None

@dataclass
class CreateChatCompletionResponseChoicesInner:
    index: Optional[int] = None
    message: Optional[ChatCompletionRequestMessage] = None
    finish_reason: Optional[str] = None

@dataclass
class CreateChatCompletionResponse:
    id: str
    object: str
    created: int
    model: str
    choices: List[CreateChatCompletionResponseChoicesInner]
    usage: Optional[CreateCompletionResponseUsage] = None

@dataclass
class ChatCompletionResponseMessage:
    role: str
    content: str

@dataclass
class CreateModerationRequest:
    input: Union[List[str], str]
    model: Optional[str] = None

@dataclass
class CreateModerationResponseResultsInnerCategories:
    hate: bool
    hate_threatening: bool
    self_harm: bool
    sexual: bool
    sexual_minors: bool
    violence: bool
    violence_graphic: bool

@dataclass
class CreateModerationResponseResultsInnerCategoryScores:
    hate: float
    hate_threatening: float
    self_harm: float
    sexual: float
    sexual_minors: float
    violence: float
    violence_graphic: float

@dataclass
class CreateModerationResponseResultsInner:
    flagged: bool
    categories: CreateModerationResponseResultsInnerCategories
    category_scores: CreateModerationResponseResultsInnerCategoryScores

@dataclass
class CreateModerationResponse:
    id: str
    model: str
    results: List[CreateModerationResponseResultsInner]

@dataclass
class CreateEmbeddingRequest:
    model: str
    input: Union[List[Any], List[int], List[str], str]
    user: Optional[str] = None

@dataclass
class CreateEmbeddingResponseDataInner:
    index: int
    object: str
    embedding: List[float]

@dataclass
class CreateEmbeddingResponseUsage:
    prompt_tokens: int
    total_tokens: int

@dataclass
class CreateEmbeddingResponse:
    object: str
    model: str
    data: List[CreateEmbeddingResponseDataInner]
    usage: CreateEmbeddingResponseUsage

if __name__ == '__main__':
    chatCompletionObj = CreateChatCompletionRequest(model= 'gpt-3.5', messages= ['a', 'b'])



"""
Conversation prompt templates.
"""

import dataclasses
from enum import auto, Enum
from typing import List, Any, Dict


class SeparatorStyle(Enum):
    """Separator styles."""

    ADD_COLON_SINGLE = auto()
    ADD_COLON_TWO = auto()
    ADD_COLON_SPACE_SINGLE = auto()
    NO_COLON_SINGLE = auto()
    ADD_NEW_LINE_SINGLE = auto()
    DOLLY = auto()
    RWKV = auto()
    PHOENIX = auto()
    ROBIN = auto()
    SINGLE = auto()



@dataclasses.dataclass
class Conversation:
    """A class that manages prompt templates and keeps all conversation history."""

    # The name of this template
    name: str
    # The system prompt
    system: str
    # Two roles
    roles: List[str]
    # All messages. Each item is (role, message).
    messages: List[List[str]]
    # The number of few shot examples
    offset: int
    # Separators
    sep_style: SeparatorStyle
    sep: str
    sep2: str = None
    # Stop criteria (the default one is EOS token)
    stop_str: str = None
    # Stops generation if meeting any token in this list
    stop_token_ids: List[int] = None
    # prepends first msg text to front of prompt
    first_msg_no_role:bool = False
    # response_prime - should we end with 'Assistant'?
    response_prime:bool = False


    

    def get_llama_2_prompt(self):
        B_INST, E_INST = "[INST]", "[/INST]"
        B_SYS, E_SYS = "<<SYS>>\n", "\n<</SYS>>\n\n"
        DEFAULT_SYSTEM_PROMPT = """You are a helpful, respectful and honest assistant. 
If a question does not make any sense, or is not factually coherent, explain why instead of answering something not correct. If you don't know the answer to a question, say so."""
        
        messages = self.messages
        #print(f'Input: {messages}')
        if messages[0][0] != "system":
            messages = [("system", '')] + messages
        messages = [[messages[1][0],B_SYS + messages[0][1] + E_SYS + messages[1][1]]] + messages[2:]
        #print(f'\nInitial rewrite: {messages}')
        ret: str = \
            ''.join([
                f"{B_INST} {(prompt[1]).strip()} {E_INST} {(answer[1]).strip()} "
                for prompt, answer in zip(
                        messages[::2],
                        messages[1::2],
                )
            ])
        #print(f'ret: {ret}')
        assert (messages[-1][0] == "user"), f"Last message must be from user, got {ret[-1][0]}"
        ret_suffix =  f"{B_INST} {(messages[-1][1]).strip()} {E_INST}"
        #print(f'ret_suffix: {ret_suffix}')
        ret=ret+ret_suffix
        #print(f'Final ret: {ret}')
        return ret

        
    def get_prompt(self, include_system=False) -> str:
        """Get the prompt for generation."""
        if self.name == 'llama-2':
            return self.get_llama_2_prompt()
        if self.sep_style == SeparatorStyle.SINGLE:
            ret = ''
            if len(self.system) > 0 and include_system:
                ret = self.system + self.sep
            for role, message in self.messages:
                if message:
                    ret += role + ": " + "<s>" + message + "</s>\n"
                else:
                    ret += role + ": " + "<s>\n"
            return ret
        if self.sep_style == SeparatorStyle.ADD_COLON_SINGLE:
            ret = ''
            if len(self.system) > 0 and include_system:
                ret = self.system + self.sep
            for role, message in self.messages:
                if message:
                    ret += role + ": " + message + self.sep
                else:
                    ret += role + ":"
            return ret
        elif self.sep_style == SeparatorStyle.ADD_COLON_TWO:
            seps = [self.sep, self.sep2]
            ret = ''
            if len(self.system) > 0 and include_system:
                ret = self.system + self.sep
            for i, (role, message) in enumerate(self.messages):
                if message:
                    ret += role + ": " + message + seps[i % 2]
                else:
                    ret += role + ":"
            return ret
        elif self.sep_style == SeparatorStyle.ADD_COLON_SPACE_SINGLE:
            ret = ''
            if len(self.system) > 0 and include_system:
                ret = self.system + self.sep
            for role, message in self.messages:
                if message:
                    ret += role + ": " + message + self.sep
                else:
                    ret += role + ": "  # must be end with a space
            return ret
        elif self.sep_style == SeparatorStyle.ADD_NEW_LINE_SINGLE:
            ret = ''
            if len(self.system) > 0 and include_system:
                ret = self.system + self.sep
            for role, message in self.messages:
                if message:
                    ret += role + "\n" + message + self.sep
                else:
                    ret += role + "\n"
            return ret
        elif self.sep_style == SeparatorStyle.NO_COLON_SINGLE:
            ret = ''
            if len(self.system) > 0 and include_system:
                ret = self.system
            for role, message in self.messages:
                if message:
                    ret += role + message + self.sep
                else:
                    ret += role
            return ret
        elif self.sep_style == SeparatorStyle.RWKV:
            ret = ''
            if len(self.system) > 0 and include_system:
                ret = self.system
            for i, (role, message) in enumerate(self.messages):
                if message:
                    ret += (
                        role
                        + ": "
                        + message.replace("\r\n", "\n").replace("\n\n", "\n")
                    )
                    ret += "\n\n"
                else:
                    ret += role + ":"
            return ret
        elif self.sep_style == SeparatorStyle.DOLLY:
            seps = [self.sep, self.sep2]
            ret = ''
            if len(self.system) > 0 and include_system:
                ret = self.system
            for i, (role, message) in enumerate(self.messages):
                if message:
                    ret += role + ":\n" + message + seps[i % 2]
                    if i % 2 == 1:
                        ret += "\n\n"
                else:
                    ret += role + ":\n"
            return ret
        elif self.sep_style == SeparatorStyle.PHOENIX:
            ret = ''
            if len(self.system) > 0 and include_system:
                ret = self.system
            for role, message in self.messages:
                if message:
                    ret += role + ": " + "<s>" + message + "</s>"
                else:
                    ret += role + ": " + "<s>"
            return ret
        elif self.sep_style == SeparatorStyle.ROBIN:
            ret = ''
            if len(self.system) > 0 and include_system:
                ret = self.system + self.sep
            for role, message in self.messages:
                if message:
                    ret += role + ":\n" + message + self.sep
                else:
                    ret += role + ":\n"
            return ret
        else:
            raise ValueError(f"Invalid style: {self.sep_style}")

    def append_message(self, role: str, message: str):
        """Append a new message."""
        self.messages.append([role, message])

    def update_last_message(self, message: str):
        """Update the last output.

        The last message is typically set to be None when constructing the prompt,
        so we need to update it in-place after getting the response from a model.
        """
        self.messages[-1][1] = message

    def to_gradio_chatbot(self):
        """Convert the conversation to gradio chatbot format."""
        ret = []
        for i, (role, msg) in enumerate(self.messages[self.offset :]):
            if i % 2 == 0:
                ret.append([msg, None])
            else:
                ret[-1][-1] = msg
        return ret

    def to_openai_api_messages(self):
        """Convert the conversation to OpenAI chat completion format."""
        ret = [{"role": "system", "content": self.system}]

        for i, (_, msg) in enumerate(self.messages[self.offset :]):
            if i % 2 == 0:
                ret.append({"role": "user", "content": msg})
            else:
                if msg is not None:
                    ret.append({"role": "assistant", "content": msg})
        return ret

    def copy(self):
        return Conversation(
            name=self.name,
            system=self.system,
            roles=self.roles,
            messages=[[x, y] for x, y in self.messages],
            offset=self.offset,
            sep_style=self.sep_style,
            sep=self.sep,
            sep2=self.sep2,
            stop_str=self.stop_str,
            stop_token_ids=self.stop_token_ids,
            first_msg_no_role=self.first_msg_no_role,
            response_prime = self.response_prime,
        )

    def dict(self):
        return {
            "template_name": self.name,
            "system": self.system,
            "roles": self.roles,
            "messages": self.messages,
            "offset": self.offset,
        }


# A global registry for all conversation templates
conv_templates: Dict[str, Conversation] = {}


def register_conv_template(template: Conversation, override: bool = False):
    """Register a new conversation template."""
    if not override:
        assert template.name not in conv_templates, f"{name} has been registered."
    conv_templates[template.name] = template


def get_conv_template(name: str) -> Conversation:
    """Get a conversation template."""
    return conv_templates[name].copy()


# A template with a one-shot conversation example
register_conv_template(
    Conversation(
        name="one_shot",
        system="A chat between a curious human and an artificial intelligence assistant. "
        "The assistant gives helpful, detailed, and polite answers to the human's questions.",
        roles=("Human", "Assistant"),
        messages=(
            (
                "Human",
                "Got any creative ideas for a 10 year old’s birthday?",
            ),
            (
                "Assistant",
                """Of course! Here are some creative ideas for a 10-year-old's birthday party:
1. Treasure Hunt: Organize a treasure hunt in your backyard or nearby park. Create clues and riddles for the kids to solve, leading them to hidden treasures and surprises.
2. Science Party: Plan a science-themed party where kids can engage in fun and interactive experiments. You can set up different stations with activities like making slime, erupting volcanoes, or creating simple chemical reactions.
3. Outdoor Movie Night: Set up a backyard movie night with a projector and a large screen or white sheet. Create a cozy seating area with blankets and pillows, and serve popcorn and snacks while the kids enjoy a favorite movie under the stars.
4. DIY Crafts Party: Arrange a craft party where kids can unleash their creativity. Provide a variety of craft supplies like beads, paints, and fabrics, and let them create their own unique masterpieces to take home as party favors.
5. Sports Olympics: Host a mini Olympics event with various sports and games. Set up different stations for activities like sack races, relay races, basketball shooting, and obstacle courses. Give out medals or certificates to the participants.
6. Cooking Party: Have a cooking-themed party where the kids can prepare their own mini pizzas, cupcakes, or cookies. Provide toppings, frosting, and decorating supplies, and let them get hands-on in the kitchen.
7. Superhero Training Camp: Create a superhero-themed party where the kids can engage in fun training activities. Set up an obstacle course, have them design their own superhero capes or masks, and organize superhero-themed games and challenges.
8. Outdoor Adventure: Plan an outdoor adventure party at a local park or nature reserve. Arrange activities like hiking, nature scavenger hunts, or a picnic with games. Encourage exploration and appreciation for the outdoors.
Remember to tailor the activities to the birthday child's interests and preferences. Have a great celebration!""",
            ),
        ),
        offset=2,
        sep_style=SeparatorStyle.ADD_COLON_SINGLE,
        sep="\n### ",
        stop_str="###",
    )
)

# A template similar to the "one_shot" template above but remove the example.
register_conv_template(
    Conversation(
        name="zero_shot",
        system="A chat between a curious human and an artificial intelligence assistant. "
        "The assistant gives helpful, detailed, and polite answers to the human's questions.",
        roles=("Human", "Assistant"),
        messages=(),
        offset=0,
        sep_style=SeparatorStyle.ADD_COLON_SINGLE,
        sep="\n### ",
        stop_str="###",
    )
)

# Vicuna v1.1 template
register_conv_template(
    Conversation(
        name="vicuna_v1.1",
        system="",
        roles=("USER", "ASSISTANT"),
        messages=(),
        offset=0,
        sep_style=SeparatorStyle.ADD_COLON_TWO,
        sep=" ",
        sep2="</s>",
        response_prime=True,
    )
)

# Koala default template
register_conv_template(
    Conversation(
        name="koala_v1",
        system="BEGINNING OF CONVERSATION:",
        roles=("USER", "GPT"),
        messages=(),
        offset=0,
        sep_style=SeparatorStyle.ADD_COLON_TWO,
        sep=" ",
        sep2="</s>",
    )
)

# Alpaca default template
register_conv_template(
    Conversation(
        name="alpaca",
        system="",
        roles=("### Instruction", "### Response"),
        messages=(),
        offset=0,
        sep_style=SeparatorStyle.ADD_COLON_SINGLE,
        sep="\n\n",
        response_prime=True,
    )
)

# OpenLLama default template (copy of alpaca)
register_conv_template(
    Conversation(
        name="openllama",
        system="",
        roles=("### Instruction", "### Response"),
        messages=(),
        offset=0,
        sep_style=SeparatorStyle.ADD_COLON_SINGLE,
        sep="\n\n",
    )
)

# Dolly V2 default template
register_conv_template(
    Conversation(
        name="dolly_v2",
        system="",
        roles=("### Instruction", "### Response"),
        messages=(),
        offset=0,
        sep_style=SeparatorStyle.DOLLY,
        sep="\n\n",
        sep2="### End",
    )
)

# ChatGLM2 default template
register_conv_template(
    Conversation(
        name="chatglm2",
        system="",
        roles=("### Instruction", "### Response"),
        messages=(),
        offset=0,
        sep_style=SeparatorStyle.DOLLY,
        sep="\n\n",
        sep2="### End",
    )
)

# OpenAssistant Pythia default template
register_conv_template(
    Conversation(
        name="oasst_pythia",
        system="",
        roles=("<|prompter|>", "<|assistant|>"),
        messages=(),
        offset=0,
        sep_style=SeparatorStyle.NO_COLON_SINGLE,
        sep="<|endoftext|>",
        response_prime=True,
    )
)

# StableLM Alpha default template
register_conv_template(
    Conversation(
        name="stablelm",
        system="",
        roles=("<|USER|>", "<|ASSISTANT|>"),
        messages=(),
        offset=0,
        sep_style=SeparatorStyle.NO_COLON_SINGLE,
        sep="",
        stop_token_ids=[50278, 50279, 50277, 1, 0],
    )
)

# Baize default template
register_conv_template(
    Conversation(
        name="baize",
        system="",
        roles=("[|Human|]", "[|AI|]"),
        messages=(
            ("[|Human|]", "Hello!"),
            ("[|AI|]", "Hi!"),
        ),
        offset=2,
        sep_style=SeparatorStyle.NO_COLON_SINGLE,
        sep="\n",
        stop_str="[|Human|]",
    )
)

# RWKV-4-Raven default template
register_conv_template(
    Conversation(
        name="rwkv",
        system="",
        roles=("Bob", "Alice"),
        messages=(
            ("Bob", "hi"),
            (
                "Alice",
                "Hi. I am your assistant and I will provide expert full response in full details. Please feel free to ask any question and I will always answer it.",
            ),
        ),
        offset=2,
        sep_style=SeparatorStyle.RWKV,
        sep="",
        stop_str="\n\n",
    )
)

# Buddy default template
register_conv_template(
    Conversation(
        name="openbuddy",
        system="",
        roles=("User", "Assistant"),
        messages=(),
        offset=0,
        sep_style=SeparatorStyle.ADD_COLON_SINGLE,
        sep="\n",
    )
)

# Phoenix default template
register_conv_template(
    Conversation(
        name="phoenix",
        system="",
        roles=("Human", "Assistant"),
        messages=(),
        offset=0,
        sep_style=SeparatorStyle.PHOENIX,
        sep="</s>",
    )
)

# ChatGPT default template
register_conv_template(
    Conversation(
        name="gpt-3.5-turbo",
        system="",
        roles=("user", "assistant"),
        messages=(),
        offset=0,
        sep_style=SeparatorStyle.NO_COLON_SINGLE,
        sep=None,
    )
)

# ChatGPT default template
register_conv_template(
    Conversation(
        name="gpt-3.5-turbo-16k",
        system="",
        roles=("user", "assistant"),
        messages=(),
        offset=0,
        sep_style=SeparatorStyle.NO_COLON_SINGLE,
        sep=None,
    )
)

# ChatGPT default template
register_conv_template(
    Conversation(
        name="gpt-4",
        system="",
        roles=("user", "assistant"),
        messages=(),
        offset=0,
        sep_style=SeparatorStyle.NO_COLON_SINGLE,
        sep=None,
    )
)

# Llama-2 default template
register_conv_template(
    Conversation(
        name="llama-2",
        system="",
        roles=("user", "assistant", "system"),
        messages=(),
        offset=0,
        sep_style=SeparatorStyle.NO_COLON_SINGLE,
        sep=None,
    )
)

# Claude default template
register_conv_template(
    Conversation(
        name="claude",
        system="",
        roles=("Human", "Assistant"),
        messages=(),
        offset=0,
        sep_style=SeparatorStyle.ADD_COLON_SINGLE,
        sep="\n\n",
    )
)

# MPT default template
register_conv_template(
    Conversation(
        name="mpt",
        system="",
        roles=("<|im_start|>user", "<|im_start|>assistant", "<|im_start|>system"),
        messages=(),
        offset=0,
        sep_style=SeparatorStyle.ADD_NEW_LINE_SINGLE,
        sep="<|im_end|>",
        sep2='',
        stop_token_ids=[50278, 0],
        response_prime=True,
    )
)
# MPT instruct template as per HG MPT 30b instruct model page
register_conv_template(
    Conversation(
        name="mpt_instruct",
        system="",
        roles=("### Instruction", "### Response",""),
        messages=(),
        offset=0,
        sep_style=SeparatorStyle.ADD_NEW_LINE_SINGLE,
        sep="\n",
        sep2='',
        stop_token_ids=[50278, 0],
        response_prime=True,
    )
)

# Bard default template
# Reference: https://github.com/google/generative-ai-python/blob/9c99bcb474a991a97a2e7d62fcdb52db7ce40729/google/generativeai/discuss.py#L150
#            https://github.com/google/generative-ai-python/blob/9c99bcb474a991a97a2e7d62fcdb52db7ce40729/google/generativeai/discuss.py#L40
register_conv_template(
    Conversation(
        name="bard",
        system="",
        roles=("0", "1"),
        messages=(),
        offset=0,
        sep_style=None,
        sep=None,
    )
)

# BiLLa default template
register_conv_template(
    Conversation(
        name="billa",
        system="",
        roles=("Human", "Assistant"),
        messages=(),
        offset=0,
        sep_style=SeparatorStyle.ADD_COLON_SPACE_SINGLE,
        sep="\n",
        stop_str="Human:",
    )
)

# RedPajama INCITE default template
register_conv_template(
    Conversation(
        name="redpajama-incite",
        system="",
        roles=("<human>", "<bot>"),
        messages=(),
        offset=0,
        sep_style=SeparatorStyle.ADD_COLON_SINGLE,
        sep="\n",
        stop_str="<human>",
    )
)

# h2oGPT default template
register_conv_template(
    Conversation(
        name="h2ogpt",
        system="",
        roles=("<|prompt|>", "<|answer|>"),
        messages=(),
        offset=0,
        sep_style=SeparatorStyle.NO_COLON_SINGLE,
        sep="</s>",
    )
)

# Robin default template
register_conv_template(
    Conversation(
        name="Robin",
        system="",
        roles=("###Human", "###Assistant"),
        messages=(),
        offset=0,
        sep_style=SeparatorStyle.ROBIN,
        sep="\n",
        stop_token_ids=[2, 396],
        stop_str="###",
    )
)

# Snoozy default template
# Reference: https://github.com/nomic-ai/gpt4all/blob/d4861030b778da6db59d21d2927a4aba4f9f1f43/gpt4all-bindings/python/gpt4all/gpt4all.py#L232
register_conv_template(
    Conversation(
        name="snoozy",
        system="",
        roles=("### Prompt", "### Response", "### Instruction:\n"),
        messages=(),
        offset=0,
        sep_style=SeparatorStyle.ADD_COLON_SINGLE,
        sep="\n",
        stop_str="###",
    )
)

# manticore default template
register_conv_template(
    Conversation(
        name="manticore",
        system="",
        roles=("USER", "ASSISTANT"),
        messages=(),
        offset=0,
        sep_style=SeparatorStyle.ADD_COLON_TWO,
        sep="\n",
        sep2="</s>",
    )
)

# Falcon default template
register_conv_template(
    Conversation(
        name="falcon",
        system="",
        roles=("User", "Assistant"),
        messages=[],
        offset=0,
        sep_style=SeparatorStyle.RWKV,
        sep="\n",
        sep2="<|endoftext|>",
        stop_str="\nUser",  # use stop_str to stop generation after stop_token_ids, it will also remove stop_str from the generated text
        stop_token_ids=[
            0,
            1,
            2,
            3,
            4,
            5,
            6,
            7,
            8,
            9,
            10,
            11,
        ],  # it better only put special tokens here, because tokenizer only remove special tokens
        response_prime=True,
    )
)

register_conv_template(
    Conversation(
        name="falcon_instruct",
        system="",
        roles=("User", "Assistant"),
        messages=(),
        offset=0,
        sep_style=SeparatorStyle.ADD_COLON_SINGLE,
        stop_str=["User:", "Assistant:"],
        stop_token_ids=[11],
        sep="\n",
        sep2="<|endoftext|>",
        first_msg_no_role=False,
        response_prime=True,
    )
)
register_conv_template(
    Conversation(
        name="falcon_GPTQ",
        system="",
        roles=("### Human", "### Assistant"),
        messages=(),
        offset=0,
        sep_style=SeparatorStyle.ADD_COLON_SINGLE,
        stop_str=["### Human"],
        stop_token_ids=[11],
        sep="\n",
        sep2="<|endoftext|>",
        first_msg_no_role=False,
    )
)
register_conv_template(Conversation(
        name="galactica",
        system="",
        roles=("Question", "Answer", ""), # third entry is system prefix. if absent defaults to user prefix.
        messages=(),
        offset=0,
        sep_style=SeparatorStyle.ADD_COLON_TWO,
        sep="\n",
        sep2="",
        response_prime=True,
    )
)
register_conv_template(Conversation(
        name="guanaco",
        system="",
        roles=("### Human", "### Assistant"),
        messages=(),
        offset=0,
        sep_style=SeparatorStyle.ADD_COLON_TWO,
        sep="\n",
        sep2="",
        response_prime=True,
    )
)
register_conv_template(Conversation(
        name="wizardLM2",
        system="",
        roles=("### Input", "### Response"),
        messages=(),
        offset=0,
        sep_style=SeparatorStyle.ADD_COLON_TWO,
        sep="\n",
        sep2="\n",
        response_prime=True,
    )
)
register_conv_template(Conversation(
        name="wizardCoder",
        system="Below is an instruction that describes a task. Write a response that appropriately completes the request",
        roles=("### Instruction", "### Response"),
        messages=(),
        offset=0,
        sep_style=SeparatorStyle.ADD_COLON_TWO,
        sep="\n",
        sep2="<|end|>",
        response_prime=True,
    )
)
register_conv_template(Conversation(
        name="wizardLM",
        system="",
        roles=("USER", "ASSISTANT"),
        messages=(),
        offset=0,
        sep_style=SeparatorStyle.ADD_COLON_TWO,
        sep="\n",
        sep2="\n",
        response_prime=True,
    )
)

# ChagGPT default template
register_conv_template(
    Conversation(
        name="polyglot_changgpt",
        system="",
        roles=("B", "A"),
        messages=(),
        offset=0,
        sep_style=SeparatorStyle.ADD_COLON_SINGLE,
        sep="\n",
    )
)

# Pygmalion template
register_conv_template(
    Conversation(
        name="pygmalion",
        system="",
        roles=("You", "Samanath"),
        messages=(),
        offset=0,
        sep_style=SeparatorStyle.ADD_COLON_SINGLE,
        sep="\n",
        response_prime=True,
    )
)

# tigerbot template
register_conv_template(
    Conversation(
        name="tigerbot",
        system="",
        roles=("### Instruction", "### Response"),
        messages=(),
        offset=0,
        sep_style=SeparatorStyle.ROBIN,
        sep="\n\n",
        stop_str="###",
    )
)
# tigerbot template
register_conv_template(
    Conversation(
        name="xgen",
        system="",
        roles=("### Human", "### Assistant"),
        messages=(),
        offset=0,
        sep_style=SeparatorStyle.ADD_COLON_SINGLE,
        sep="\n\n",
        stop_str="###",
    )
)


if __name__ == "__main__":
    conv = get_conv_template("llama-2")
    conv.append_message('system', "You are a friendly AI")
    conv.append_message(conv.roles[0], "Hi!")
    conv.append_message(conv.roles[1], "How are you?")
    conv.append_message(conv.roles[0], "What is your name")
    print(conv.get_prompt())



import torch
import transformers
from transformers import AutoModelForCausalLM, AutoTokenizer
from alphawave_pyexts import serverUtils as sv
from accelerate import load_checkpoint_and_dispatch
import datetime
import os
import requests, socket
from queue import Queue
import json
import sys
import time
import traceback


model_name = "TheBloke/Wizard-Vicuna-30B-Uncensored-fp16"
print(f'**** Loading {model_name}')

if __name__ == '__main__':

    tokenizer = AutoTokenizer.from_pretrained(model_name, use_fast=False, skip_special_tokens=True)
    model = AutoModelForCausalLM.from_pretrained(
        model_name,
        load_in_8bit=True,
        device_map="auto",
    )
    model.tie_weights()
    pipeline = transformers.pipeline(
        "text-generation",
        model=model,
        tokenizer=tokenizer,
        device_map="auto",
    )
    
    print(f'server started {model_name} on port 5004, use vicuna_v1.1')

    #sv.server(model=model, tokenizer=tokenizer, stop_str=['###', '### Input', '### Response'])
    sv.server(tokenizer=tokenizer, pipeline=pipeline, stop_str=['USER', 'ASSISTANT', '<\s>', '<s>'])



import torch
import transformers
from transformers import (
    AutoTokenizer,
    BitsAndBytesConfig,
    AutoModelForCausalLM,
)
from alphawave_pyexts import serverUtils as sv


model_name = '/home/bruce/Downloads/llama/llama-2-70b'
print(f"Loading {model_name}")

model = AutoModelForCausalLM.from_pretrained(model_name, torch_dtype=torch.float16, load_in_4bit=True, device_map="auto", trust_remote_code=True)
tokenizer = AutoTokenizer.from_pretrained(model_name)

text = "Hello my name is"
device = "cuda:0"

pipeline = transformers.pipeline(
    "text-generation",
    model=model,
    tokenizer=tokenizer,
    trust_remote_code=True,
    device_map="auto",
)
print(f"server started {model_name} into memory, use alpaca")

sv.server(model=None, tokenizer=tokenizer, pipeline = pipeline, stop_str=['User:'])
#sv.server(model=model, tokenizer=tokenizer, pipeline = None, stop_str=['User:'])



from pyee import AsyncIOEventEmitter
from typing import Dict, Any, List, Optional
from promptrix.promptrixTypes import PromptFunctions, PromptMemory, Tokenizer, Message, RenderedPromptSection
from promptrix.PromptSectionBase import PromptSectionBase
from alphawave_agents.agentTypes import Command
import traceback

class AgentCommandSection(PromptSectionBase):
    def __init__(self, commands: Dict[str, Command], tokens: int = -1, required: bool = True, one_shot=False, syntax = 'JSON'):
        super().__init__(tokens, required)
        self._commands = commands
        self.tokens = tokens
        self.required = required
        self.one_shot = one_shot
        self.syntax = syntax
        
    def renderAsMessages(self, memory: Any, functions: Any, tokenizer: Any, max_tokens: int) -> Dict[str, Any]:
        # Render commands to message content
        content = 'You have the following  commands available in addition to known facts and reasoning:\n'
        for command in self._commands.values():
            content += f'\t{command.title}:\n'
            content += f'\t\tuse: {command.description}\n'
            #if command.inputs:
            #    content += f'\t\tinputs: {command.inputs}\n'
            #if command.output:
            #    content += f'\t\toutput: {command.output}\n'
            #if self.one_shot:
            content += command.one_shot(self.syntax)
        # Return as system message
        result = None
        try:
            length = len(tokenizer.encode(content))
            result = self.return_messages([{'role': 'system', 'content': content}], length, tokenizer, max_tokens)
        except Exception as e:
            traceback.print_exc()
        return result




from typing import Any, Dict, Optional
from alphawave_agents.SchemaBasedCommand import SchemaBasedCommand
from alphawave_agents.SchemaBasedCommand import CommandSchema as sbcCommandSchema
from alphawave_agents.agentTypes import TaskResponse
from dataclasses import dataclass, asdict
import traceback

@dataclass
class CommandSchema(sbcCommandSchema):
    schema_type: str
    title: str
    description: str
    properties: Dict[str,Dict[str,str]]
    required: list[str]
    returns: str

@dataclass
class MathCommandInput:
    code:str

schema = CommandSchema(
    schema_type="object",
    title="math",
    description="evaluate a mathematical expression to calculate its value",
    properties={
        "code": {
            "type": "string",
            "description": "expression to evaluate"
        }
    },
    required=["code"],
    returns="the calculated value"
)

class MathCommand(SchemaBasedCommand):

    def __init__(self, title = None, description = None):
        super().__init__(schema, title, description)

    def execute(self, input: MathCommandInput, memory: Any, functions: Any, tokenizer: Any) -> Any:
        try:
            exp = input['code']
            exp = exp.replace('\\*', '*')
            exp = exp.replace('\\+', '+')
            exp = exp.replace('\\-', '-')
            exp = exp.replace('\\/', '/')
            return eval(exp)
        except Exception as err:
            message = str(err)
            return asdict(TaskResponse('TaskResponse', 'error', message))




import toml
from toml import TomlDecodeError
from cerberus import Validator
from promptrix.promptrixTypes import Message, PromptFunctions, PromptMemory, Tokenizer
from alphawave.alphawaveTypes import PromptResponse, Validation, PromptResponseValidator
from pyee import AsyncIOEventEmitter
import sys
import traceback
import re


class TOMLResponseValidator(PromptResponseValidator):
    def __init__(self, schema=None, missing_toml_feedback="Response was not formatted as expected. "):
        self.schema = schema
        self.missing_toml_feedback = missing_toml_feedback
        self.validator = Validator(schema)
        try:
            self.feedback_schema = self.extract_toml_template(schema)
        except Exception as e:
            traceback.print_exc()
        #print(f'***** TOMLResponseValidator feedback {self.feedback_schema}')
        
    def find_toml(self, s):
        if s is None:
            return s
        # Look for [RESPONSE] ... [STOP]

        s = re.sub('\n+', '\n', s)
        cleaned_s = ""
        for char in s:
            if ord(char) >= 10:
                cleaned_s += char
        s = cleaned_s
        s = s.replace('\\*', '*')
        s = s.replace('\\+', '+')
        s = s.replace('\\-', '-')
        s = s.replace('\\/', '/')
        s = s.replace('RESPONSE:', '[RESPONSE]')
        start = s.find('[RESPONSE]')
        if start < 0 and self.schema is None:
            #print(f'***** find_toml no start')
            return ''
        elif start < 0:
            #print(f'looking for {list(self.schema.keys())[0]}')
            start = s.find(list(self.schema.keys())[0])
            if start < 0:
                #print(f'***** find_toml no start')
                return ''
        end = s[start:].find('[STOP]')
        if end < 0:
            end = (s[start+1:]).find('[')
        if end < 0:
            #print(f'***** find_toml {start} no end\n{s[start:]}')
            return s[start:]
        else:
            #print(f'***** find_toml {start}, {end}\n{s[start:start+end-1]}')
            return s[start:start+end-1]

    def extract_toml_template(self,schema, prefix=None):
        if prefix is None:  ### top level
            #print(f'\n***** template extract input\n{schema}')
            template = '[RESPONSE]\n'
        else:
            template = ''
        if schema is not None and type(schema) == dict:
            for item in schema:
                value = schema[item]
                if prefix is None:
                    key = item
                else:
                    key = prefix+"."+item
                if type(value) == dict:
                    if 'type' in value and value['type'] == 'dict':
                        #print(f'\n***** template dict {key},{value}')
                        # recurse on nested structure
                        if 'schema' in value:
                            subtemplate = self.extract_toml_template(value['schema'], key)
                            template = template+subtemplate
                        elif 'keysrules' in value:
                            template = template+key+"={}\n"
                    elif 'meta' in value:
                        template = template+key+'="'+value['meta']+'"\n'
                    elif 'type' in value:
                        template = template+key+'="'+value['type']+'"\n'
                else:
                    template = template+key+'='+value+'\n'
                    
        if prefix is None:
            template += '[STOP]\n'
        #print(f'\n***** template extract \n{template}')
        return template
                
    def validate_response(self, memory: PromptMemory, functions: PromptFunctions, tokenizer: Tokenizer, response: PromptResponse, remaining_attempts: int) -> Validation:
        message = response['message']
        
        text = message if isinstance(message, str) else message.get('content', '')
        #print(f'***** TOMLResponseValidator raw input type {type(text)}\n{text}\n')
        try:
            toml_extract = self.find_toml(text)
        except Exception as e:
            traceback.print_exc()
        #print(f'\n***** TOMLResponseValidator toml_extract \n{toml_extract}\n')
        if toml_extract is None or len(toml_extract) < 5:
            #print(f'***** TOMLResponseValidator failure no toml', file=sys.stderr)
            return {
                'type': 'Validation',
                'valid': False,
                'feedback': self.missing_toml_feedback+f' using this template:\n{self.feedback_schema}',
            }

        # Validate the response against the schema
        errors = None
        try:
            response_as_dict = toml.loads(toml_extract)
            if 'RESPONSE' in list(response_as_dict.keys()):
                response_as_dict = response_as_dict['RESPONSE']
            #print(f'\n***** TOMLResponseValidator as_dict \n{response_as_dict}\n')
        except TomlDecodeError as e:
            #print(f'***** TOMLResponseValidator TomlDecodeError {str(e)}\n{toml_extract}')
            return {
                'type': 'Validation',
                'valid': False,
                'feedback': f'Response found but not a complete TOML form. e fixes:\n{str(e)}. respond using this format:\n{self.feedback_schema} '
            }

        if self.schema:
            try:
                #print(f'***** TOMLResponseValidator validate input against \n{self.schema}')
                v = Validator(self.schema)
                validation_result = v.validate(response_as_dict)
                if validation_result:
                    #print(f'***** TOMLResponseValidator validation passed!{response_as_dict}')
                    return {
                        'type': 'Validation',
                        'valid': True,
                        'value': response_as_dict
                    }
                else:
                    #print(f'***** TOMLResponseValidator schema validation failed {v._errors}\n{response_as_dict}\n')
                    return {
                        'type': 'Validation',
                        'valid': False,
                        'feedback': f'Response TOML found but didnt match schema. Apply these fixes:\n{v._errors} response template:\n{self.feedback_schema}\n'
                    }
            except Exception as e:
                #print(f'***** TOMLResponseValidator validator exception {str(e)}')
                return {
                    'type': 'Validation',
                    'valid': False,
                    'feedback': f'TOML form found but exception attempting to validate. Repair and respond using this template:\n{self.feedback_schema} '
                }      
    
        else:
            # Return the last object
            #print(f'***** TOMLResponseValidator exit last object {response_as_dict}')
            return {
                'type': 'Validation',
                'valid': True,
                'value': response_as_dict
            }




import requests
from dataclasses import dataclass, asdict
from typing import Optional, Dict, Any, Union
import time
import json
import asyncio
from promptrix.promptrixTypes import PromptFunctions, PromptMemory, PromptSection, Tokenizer
from promptrix.SystemMessage import SystemMessage
from promptrix.ConversationHistory import ConversationHistory
from promptrix.AssistantMessage import AssistantMessage

from alphawave.alphawaveTypes import PromptCompletionClient, PromptCompletionOptions, PromptResponse
from alphawave.internalTypes import ChatCompletionRequestMessage, CreateChatCompletionRequest, CreateChatCompletionResponse, CreateCompletionRequest, CreateCompletionResponse
from alphawave.Colorize import Colorize
import alphawave_pyexts.utilityV2 as ut
import alphawave_pyexts.LLMClient as client

@dataclass
class OSClientOptions(PromptCompletionOptions):
    def __init__(self, apiKey, model='wizardLM', organization = None, endpoint ='127.0.0.1', port=5004,logRequests =False):
        self.apiKey = apiKey
        self.organization = organization
        self.endpoint = endpoint
        self.port = port
        self.logRequests = logRequests
        self.model = model
        
def update_dataclass(instance, **kwargs):
    for key, value in kwargs.items():
        if hasattr(instance, key):
            setattr(instance, key, value)
    return instance

def get_values(instance, keys):
    values = []
    for key in keys:
        if hasattr(instance, key):
            values.append(getattr(instance, key))
        else:
            values.append(None)
    return values

def display_options(options):
    options_dict = options.__dict__
    print()
    for item in options_dict.keys():
        print(item, options_dict[item])
    print()
    
@dataclass
class Response:
    status_code: int
    text: str
    headers: Dict[str,str] = None
    reason: str = ''
    
class OSClient(PromptCompletionClient):
    DefaultEndpoint = 'https://api.openai.com'
    UserAgent = 'AlphaWave'

    def __init__(self, **kwargs):
        self.options = OSClientOptions(apiKey=None, organization=None, endpoint= '127.0.0.1', port = 5004, logRequests=False)
        update_dataclass(self.options, **kwargs)
        if self.options.endpoint:
            self.options.endpoint = self.options.endpoint.strip()
            if self.options.endpoint.endswith('/'):
                self.options.endpoint = self.options.endpoint[:-1]

        self._session = requests.Session()

    def completePrompt(self, memory: PromptMemory, functions: PromptFunctions, tokenizer: Tokenizer, prompt: PromptSection, options: PromptCompletionOptions) -> PromptResponse:
        if isinstance(options, dict):
            argoptions = options
            options = PromptCompletionOptions(completion_type = argoptions['completion_type'],
                                              model = argoptions['model'],
                                              max_input_tokens = argoptions['max_input_tokens'],
                                              temperature = argoptions['temperature'],
                                              top_p = argoptions['top_p'],
                                              max_tokens = argoptions['max_tokens'],
                                              stop = argoptions['stop'],
                                              presence_penalty = argoptions['presence_penalty'],
                                              frequency_penalty = argoptions['frequency_penalty']
                                              )
        startTime = time.time()
        max_input_tokens = 1500
        if hasattr(options, 'max_input_tokens') and getattr(options, 'max_input_tokens') is not None:
            max_input_tokens = options.max_input_tokens
        
        result = prompt.renderAsMessages(memory, functions, tokenizer, max_input_tokens)
        
        if result.tooLong:
            return {'status': 'too_long', 'message': f"The generated chat completion prompt had a length of {result.length} tokens which exceeded the max_input_tokens of {max_input_tokens}."}
        if self.options.logRequests:
            print(Colorize.title('CHAT PROMPT:'))
            for msg in result.output:
                if not isinstance(msg, dict):
                    print(Colorize.output(msg))
                    msg = msg.__dict__
                print(Colorize.output(json.dumps(msg, indent=2)), end='')
            print()

        request = self.copyOptionsToRequest(CreateChatCompletionRequest(model = options.model, messages =  result.output), options, ['max_tokens', 'temperature', 'top_p', 'n', 'stream', 'logprobs', 'echo', 'stop', 'presence_penalty', 'frequency_penalty', 'best_of', 'logit_bias', 'user'])

        response = self.createChatCompletion(request)
        if self.options.logRequests:
            print(Colorize.title('CHAT RESPONSE:'))
            print(Colorize.value('status', response.status))
            print(Colorize.value('duration', time.time() - startTime, 'ms'))
            print(Colorize.output(response.message))
        
        if response.status == 'success':
            completion = response.message
            return {'status': 'success', 'message': completion}
        else:
            return {'status': 'error', 'message': f"The chat completion API returned an error status of {response.status}: {response.message}"}

    def addRequestHeaders(self, headers: Dict[str, str], options: OSClientOptions):
        headers['Authorization'] = f"Bearer {options.apiKey}"
        if options.organization:
            headers['OS-Organization'] = options.organization

    def copyOptionsToRequest(self, target: Dict[str, Any], src: Any, fields: list) -> Dict[str, Any]:
        for field in fields:
            if hasattr(src, field) and getattr(src, field) is not None:
                setattr(target,field, getattr(src,field))
        return target

    def createChatCompletion(self, request: CreateChatCompletionRequest) -> requests.Response:
        url = f"{self.options.endpoint or self.DefaultEndpoint}/v1/chat/completions"
        requestHeaders = {
            'Content-Type': 'application/json',
            'User-Agent': self.UserAgent
        }
        result = ''
        try:
            result = ut.ask_LLM(request.model,
                                request.messages,
                                request.max_tokens,
                                request.temperature,
                                request.top_p,
                                self.options.endpoint,
                                self.options.port
                                )
        except Exception as e:
            return PromptResponse(status='error',message=str(e))
        return PromptResponse(status='success', message=result)



from typing import Any, Dict, List
from colorama import Fore, Style
from dataclasses import dataclass, asdict
import asyncio
import copy
from promptrix.SystemMessage import SystemMessage
from promptrix.VolatileMemory import VolatileMemory
from alphawave_agents.agentTypes import TaskResponse
from alphawave_agents.SchemaBasedCommand import SchemaBasedCommand
from alphawave_agents.SchemaBasedCommand import CommandSchema as sbcCommandSchema
from alphawave_pyexts.llmsearch import search_service
import json

@dataclass
class CommandSchema(sbcCommandSchema):
    schema_type: str
    title: str
    description: str
    properties: Dict[str,Dict[str,str]]
    required: list[str]
    returns: str


search_schema = CommandSchema(
    schema_type="object",
    title="search",
    description="web search for provided query",
    properties={
        "query": {
            "type": "string",
            "description": "web query string"
            }
    },
    required=['query'],
    returns='search result'
)


class SearchCommand(SchemaBasedCommand):
    def __init__(self, client, model, title=None, name=None, description=None, return_urls=True, logResponse=False, max_chars=1500):
        super().__init__(search_schema, title, description)
        self.client = client
        self.model = model
        self.return_urls = return_urls
        self.max_chars = max_chars
        self.logResponse = logResponse
        
    def execute(self, input: input, memory: Any, functions: Any, tokenizer: Any) -> Any:
        try:
            print(f' starting web search {input}')
            if type(input) == dict and 'query' in input:
                query = input['query']
                memory = VolatileMemory()  # don't let anything bleed back to surrounding task
                response = search_service.run_chat(self.client, query, self.model, memory, functions, tokenizer, self.max_chars)
                sc_text = ''
                sc_urls = []
                if type(response) is list:
                    for item in response:
                        if type(item) == dict and 'url' in item:
                            sc_urls.append('\n'+item['url'])
                        if type(item) == dict and 'text' in item:
                            sc_text += '\n'+(item['text'])
                    sc_text = sc_text[:max(len(sc_text)-1, self.max_chars)]
                if self.return_urls:
                    #print(f' search returning search results')
                    return {'type':'TaskResponse','status':'success', 'message':sc_text+'\n'+str(sc_urls)}
                else:
                    return {'type':'TaskResponse','status':'success', 'message':sc_text}
                    
        except Exception as err:
            message = str(err)
            #print(f' search returning {message}')
            return asdict(TaskResponse('TaskResponse', 'error', message))



from pyee import AsyncIOEventEmitter
import asyncio
from typing import Dict, Any, Union
from promptrix.promptrixTypes import Message, PromptFunctions, PromptSection, PromptMemory, Tokenizer
from promptrix.ConversationHistory import ConversationHistory
from promptrix.AssistantMessage import AssistantMessage
from promptrix.UserMessage import UserMessage
from promptrix.FunctionRegistry import  FunctionRegistry
from promptrix.GPT3Tokenizer import GPT3Tokenizer
from promptrix.Prompt import Prompt
from promptrix.VolatileMemory import VolatileMemory
from promptrix.Utilities import Utilities
from alphawave.alphawaveTypes import PromptCompletionOptions
from alphawave.JSONResponseValidator import JSONResponseValidator
from alphawave.TOMLResponseValidator import TOMLResponseValidator
from alphawave.Response import Response
from alphawave.AlphaWave import AlphaWave, AlphaWaveOptions
from alphawave.MemoryFork import MemoryFork
from alphawave_agents.agentTypes import Command, AgentThought, AgentThoughtSchemaJSON, AgentThoughtSchemaTOML
from alphawave_agents.MathCommand import MathCommand
import traceback


class AgentCommandValidator:
    def __init__(self, commands: dict[str, Command], client=None, model: str='', syntax:str='JSON',
                 memory=VolatileMemory(), history_variable='history'):
        self._model = model
        self._client = client
        self._syntax = syntax
        self._memory=memory
        self._history_variable = history_variable
        
        if syntax == 'JSON':
            self._input_validator = JSONResponseValidator(AgentThoughtSchemaJSON, 'No valid command was found in the response. Respond using a completed instance of the specified template for the next command.')
        else:
            #print(f"***** AgentCommandValidator creating TOML")
            self._input_validator = TOMLResponseValidator(AgentThoughtSchemaTOML, 'No valid command was found. Respond using an instantiation of the specified template for the next command.')

        self._commands = commands

    def validate_response(self, memory, functions, tokenizer, response, remaining_attempts) -> Union[AgentThought, None]:
        try:
          #print(f"***** AgentCommandValidator validate syntax: \n{self._syntax}\nresponse\n{response}")
          message = response['message']
          raw_text = message if isinstance(message, str) else message.get('content', '').replace('\n','')
          if (self._syntax == 'JSON' and '{"command":' not in raw_text) or (self._syntax == 'TOML' and 'response' not in raw_text.lower()):
              # means no form found, assume llm doesn't need to use a command
              #print(f"***** AgentCommandValidator no command found")
              return {
                  'type': 'Validation',
                  'valid': True,
                  'value': raw_text,
                  'feedback': 'no command'
              }
          
          # Validate that the response contains a thought
          validation_result = self._input_validator.validate_response(memory, functions, tokenizer, response, remaining_attempts)
          if not validation_result['valid']:
              #print(f'*****AgentCommandValidator validate failed {validation_result}')
              return validation_result

          # Validate that the command exists
          thought = validation_result['value']
          #print(f'*****AgentCommandValidator post validate thought  \n{thought}\n')
          if not('command' in thought) or not('inputs' in thought):
              #print(f"***** AgentCommandValidator command or inputs not found")
              return {
                  'type': 'Validation',
                  'valid': False,
                  'feedback': f'command not found or invalid, or inputs missing. The commands you have are {list(self._commands.keys())}'
              }

          command_name = thought['command']
          if command_name not in self._commands:
              #print(f"***** AgentCommandValidator no such command {command_name}")
              return {
                  'type': 'Validation',
                  'valid': False,
                  'feedback': f'The command {command_name} does not exist. The only commands you have are {list(self._commands.keys())}'
              }
          
          # Validate that the command input is valid
          command = self._commands[command_name]
          command_validation_result = command.validate(thought['inputs'] or {}, memory, functions, tokenizer, syntax = self._syntax)
          if command_validation_result['valid']:
              #print(f"***** AgentCommandValidator command validation success\n{command_validation_result}\n")
              #validation_result['value']['inputs'] = command_validation_result['value']
              return validation_result
          else:
            return {
                'type': 'Validation',
                'valid': False,
                'feedback': 'The command '+thought['command'] +' inputs were missing or malformed. Provide them in this format:\n'
                +command.one_shot(self._syntax)+'\n, substituting actual values for command inputs\n'
                }

        except Exception as e:
            traceback.print_exc()
            return {
                'type': 'Validation',
                'valid': False,
                'feedback': f'The command validation failed. try again {str(e)}'
                }
        #print(f"***** AgentCommandValidator generic exit fail {thought['command']['name']}")
        return {
            'type': 'Validation',
            'valid': False,
            'feedback': f'The command validation failed. try again {str(e)}'
                }

    ### interesting experiment, set aside for now
    def repair_args(self, command, fail_thought):
        #print(f"***** AgentCommandValidator recovery attempt keys {list(self._memory._memory.keys())}")
        args_validator = JSONResponseValidator(command.schema, "invalid command inputs syntax, use: {command.one_shot()\n}")
        fork = MemoryFork(self._memory)
        prompt_options=PromptCompletionOptions(completion_type='chat', model=self._model, temperature=0.1)
        args_prompt=Prompt([ConversationHistory(self._history_variable),
                            UserMessage(f'invalid command args: {fail_args}, repair using this format: {command.schema["properties"]}')])
        #print(f"***** AgentCommandValidator recovery attempt wave built\ninvalid command args: {fail_args}, repair using this format: {command.schema['properties']}")
        wave = AlphaWave(client=self._client, prompt=args_prompt, prompt_options=prompt_options, memory=fork)
        #print(f"***** AgentCommandValidator recovery attempt wave built")
        args = None
        try:
            args = wave.completePrompt()
            #print(f"***** AgentCommandValidator recovery wave result {args}")
        except Exception as e:
            traceback.print_exc()
        #print(f'***** recovery result {args}')
        if args:
            return args
        #print(f"***** AgentCommandValidator recovery returning fail_args {fail_args}")
        return fail_args


from alphawave.OSClient import OSClient




from typing import Union
from promptrix.promptrixTypes import PromptFunctions, PromptMemory, PromptSection, Tokenizer
from  promptrix.promptrixTypes import Message
from alphawave.alphawaveTypes import PromptCompletionClient, PromptCompletionOptions, PromptResponse, PromptResponseStatus
import json

class TestClient(PromptCompletionClient):
    def __init__(self, status: PromptResponseStatus = 'success', response: Union[str, Message] = {'role': 'assistant', 'content': "Hello World"}):
        self.status = status
        self.response = response

    async def completePrompt(self, memory: PromptMemory, functions: PromptFunctions, tokenizer: Tokenizer, prompt: PromptSection, options: PromptCompletionOptions) -> PromptResponse:
        return {'status': self.status, 'message': self.response}



from dataclasses import dataclass
from typing import Any, Optional, Union, List
import json

# Equivalent to TypeScript's import statement
from promptrix.promptrixTypes import Message, PromptFunctions, PromptMemory, PromptSection, Tokenizer

class PromptCompletionClient:
    def completePrompt(self, memory: PromptMemory, functions: PromptFunctions, tokenizer: Tokenizer, prompt: PromptSection, options: 'PromptCompletionOptions') -> 'Promise[PromptResponse]':
        pass

class PromptResponseValidator:
    def validate_response(self, memory: PromptMemory, functions: PromptFunctions, tokenizer: Tokenizer, response: 'PromptResponse', remaining_attempts: int) -> 'Promise[Validation]':
        pass

@dataclass
class PromptCompletionOptions:
    completion_type: str  = 'chat' # 'text' | 'chat'
    model: str = ''
    max_input_tokens: int =8000
    temperature: float = 0.7
    top_p: float = 1.0
    max_tokens: int = 500
    stop: str = None
    #presence_penalty: float = 1.0
    #frequency_penalty: float = 1.0
    #logit_bias = None
    #best_of = None

PromptResponseStatus = ['success', 'error', 'rate_limited', 'invalid_response', 'too_long']

@dataclass
class PromptResponse:
    status: str  # PromptResponseStatus
    message: Union[Message, str]

@dataclass
class Validation:
    type: str  # 'Validation'
    valid: bool
    feedback = None
    value = None



import requests, time, copy
from typing import Optional, Dict, Any, Union
from dataclasses import dataclass, asdict
import json, os
from promptrix.promptrixTypes import PromptFunctions, PromptMemory, PromptSection, Tokenizer
from promptrix.SystemMessage import SystemMessage
from promptrix.ConversationHistory import ConversationHistory
from promptrix.AssistantMessage import AssistantMessage

from alphawave.alphawaveTypes import PromptCompletionClient, PromptCompletionOptions, PromptResponse
from alphawave.internalTypes import ChatCompletionRequestMessage, CreateChatCompletionRequest, CreateChatCompletionResponse, CreateCompletionRequest, CreateCompletionResponse
from alphawave.Colorize import Colorize
import openai
import asyncio

openai.api_key = os.getenv("OPENAI_API_KEY")

@dataclass
class OpenAIClientOptions:
    def __init__(self, apiKey=None, organization = None, endpoint = None, logRequests = False):
        self.apiKey = apiKey
        self.organization = organization
        self.endpoint = endpoint
        self.logRequests = logRequests

def update_dataclass(instance, **kwargs):
    for key, value in kwargs.items():
        if hasattr(instance, key):
            setattr(instance, key, value)

class OpenAIClient(PromptCompletionClient):
    DefaultEndpoint = 'https://api.openai.com'
    UserAgent = 'AlphaWave'

    def __init__(self, **kwargs):
        self.options = {'apiKey':None, 'organization':None, 'endpoint':None, 'logRequests':False}
        self.options.update(kwargs)
        if not self.options['apiKey']:
            print("Client created without an 'apiKey'.")
            raise ValueError

        self._session = requests.Session()

    def completePrompt(self, memory: PromptMemory, functions: PromptFunctions, tokenizer: Tokenizer, prompt: PromptSection, options: PromptCompletionOptions) -> PromptResponse:
        startTime = time.time()
        max_input_tokens = 1024

        #
        ### form prompt messages
        #
        if isinstance(options, dict):
            argoptions = options
            options = PromptCompletionOptions(completion_type = argoptions['completion_type'], model = argoptions['model'])
            update_dataclass(options, **argoptions)
        if hasattr(options, 'max_input_tokens') and getattr(options, 'max_input_tokens') is not None:
            max_input_tokens = options.max_input_tokens
        result = prompt.renderAsMessages(memory, functions, tokenizer, max_input_tokens)
        if result.tooLong:
            return {'status': 'too_long', 'message': f"Prompt length of {result.length} tokens exceeds max_input_tokens of {max_input_tokens}."}

        #
        ### form json request message
        #
        request = self.copyOptionsToRequest(
            CreateChatCompletionRequest(model=options.model, messages=result.output),
            options,
            ['max_tokens', 'temperature', 'top_p', 'n', 'stream', 'logprobs', 'echo', 'stop', 'presence_penalty', 'frequency_penalty', 'best_of', 'logit_bias', 'user'])
        jsonbody = asdict(request)
        keys = list(jsonbody.keys())
        for key in keys:
            if jsonbody[key] is None:
                del jsonbody[key]
        
        if self.options['logRequests']:
            print(Colorize.title('CHAT PROMPT:'))
            for msg in result.output:
                if not isinstance(msg, dict):
                    msg = msg.__dict__
                print(Colorize.output(json.dumps(msg, indent=2)), end='')
            print()
            for key in list(jsonbody.keys()):
                if key != 'messages':
                    print(Colorize.value(key, jsonbody[key]), end=',')
            print()

        result = openai.ChatCompletion.create(**jsonbody)

        if self.options['logRequests']:
            print(Colorize.title('CHAT RESPONSE:'))
            print(Colorize.value('duration', time.time() - startTime, 'ms'))
            print(Colorize.value('usage', result['usage']))
            print(Colorize.value('result message', result['choices'][0]['message']))
            #print(Colorize.output(response.json()))
        return {'status': 'success', 'message': str(result['choices'][0]['message']['content'])}

        """
        if response.status_code < 300:
            completion = response.json().get('choices')[0]
            return {'status': 'success', 'message': completion.get('message', {'role': 'assistant', 'content': ''})}
        elif response.status_code == 429:
            if self.options['logRequests']:
                print(Colorize.title('HEADERS:'))
                print(Colorize.output(response.headers))
            return {'status': 'rate_limited', 'message': 'The chat completion API returned a rate limit error.'}
        elif response.status_code == 503:
            if self.options['logRequests']:
                print(Colorize.title('HEADERS:'))
                print(Colorize.output(response.headers))
            return {'status': 'server unavailable', 'message': 'The chat completion API returned server unavailable or overloaded.'}
        else:
            return {'status': 'error', 'message': f"The chat completion API returned an error status of {response.status_code}: {response.reason}"}
        """
        
    def addRequestHeaders(self, headers: Dict[str, str], options: OpenAIClientOptions):
        headers['Authorization'] = f"Bearer {options['apiKey']}"
        if options['organization']:
            headers['OpenAI-Organization'] = options['organization']

    def copyOptionsToRequest(self, target: Dict[str, Any], src: Any, fields: list) -> Dict[str, Any]:
        for field in fields:
            if hasattr(src, field) and getattr(src, field) is not None:
                setattr(target,field, getattr(src,field))
        return target

    def createChatCompletion(self, request: CreateChatCompletionRequest) -> requests.Response:
        return self.post(request)




import argparse
import json
import traceback
import socket
import requests
import re
from alphawave_pyexts.conversation import Conversation, SeparatorStyle
import alphawave_pyexts.conversation as cv
import tkinter as tk
import time

MODEL_NAME = None; WORKER_ADDR=None; CONTROLLER_ADDRESS = "http://localhost:21001"
USER_PREFIX = 'user'
ASSISTANT_PREFIX = 'assistant'
SYSTEM_PREFIX = 'system'

host='192.168.1.195'
port = 5004


def get_available_models():
    return list(cv.conv_templates.keys())

def run_query(model, messages, max_tokens, temp, top_p, host = host, port = port, tkroot = None, tkdisplay=None, format=True): 
    global USER_PREFIX, ASSISTANT_PREFIX, SYSTEM_PREFIX

    conv=cv.get_conv_template(model)
    user_prompt = conv.roles[0]
    asst_prompt = conv.roles[1]
    system_prompt = conv.roles[1]
    
    USER_PREFIX = conv.roles[0]
    ASSISTANT_PREFIX = conv.roles[1]
    SYSTEM_PREFIX = conv.system
    
    # set this so client can check for run-on, although this test should pbly be here!
    if format:
        prime=''
        for msg_idx, msg in enumerate(messages):
            #print(msg_idx, msg)
            role = msg['role']
            ### conv.system is a prompt msg, and will be inserted as the first entry by conv.get_prompt()
            if role.lower() == 'system' and msg_idx==0:
                if len(conv.system)>0:
                    prime = conv.system+'\n'+msg['content']
                else:
                    prime = msg['content']
                if len(conv.roles)>2:
                    conv.append_message(conv.roles[2], prime)
                    prime=''
            elif role.lower() == 'user' or role.lower() == 'system' or role == conv.roles[0]:
                role_index = 0
                conv.append_message(conv.roles[role_index], msg['content'])
            else:
                role_index = 1
                conv.append_message(conv.roles[role_index], msg['content'])
                #print(conv.messages[-1])

        # priming prompt
        if conv.response_prime:
            conv.append_message(conv.roles[1], '')

        prompt = conv.get_prompt()
        if len(prime) > 0:
            prompt = prime+conv.sep+prompt
    else:
        prompt = messages
    prompt = re.sub('\n{3,}', '\n\n', prompt)
    #print(f'***** llm output prompt string {prompt}')
    server_message = {'prompt':prompt, 'temp': temp, 'top_p':top_p, 'max_tokens':max_tokens, 'user_prompt':USER_PREFIX}
    smj = json.dumps(server_message)
    try:
        client_socket = socket.socket()  # instantiate
        client_socket.connect((host, port))  # connect to the server
        client_socket.settimeout(240)
        server_message = {'prompt':prompt, 'temp': temp, 'top_p':top_p, 'max_tokens':max_tokens, 'user':user_prompt, 'asst':asst_prompt}
        smj = json.dumps(server_message)
        client_socket.sendall(smj.encode('utf-8'))
        client_socket.sendall(b'x00xff')
        response = ''
        while True:
            s = client_socket.recv(1024) # break every 32 chars to stream output
            if s is None or not s:
                break
            if (len(s) > 5 and s[-3:] == b'xff' and s[-6:-3] == b'x00'):
                s = s[:-6].decode('utf-8')
                s = s.replace('</s>', '')
                s = s.replace('<s>', '')
                if tkdisplay is not None:
                    tkdisplay.insert(tk.END, s)
                    if tkroot is not None:
                        tkroot.update()
                response += s
                break
            else:
                s = s.decode('utf-8')
                s = s.replace('</s>', '')
                s = s.replace('<s>', '')
                if tkdisplay is not None:
                    tkdisplay.insert(tk.END, s)
                    if tkroot is not None:
                        tkroot.update()
                response += s
                # not sure why sep2 code below is here???
                if conv.sep2 is not None:
                    sep2_idx = response.find(conv.sep2)
                    if sep2_idx >= 0:
                        response = response[sep2_idx:]
                        break
        client_socket.close()  # close the connection
        #check for run on hallucination in response 
        runon_idx = response.find(conv.roles[0])
        if runon_idx > 0:
            response = response[:runon_idx]
        return response
    except: 
        traceback.print_exc()
    return ''
    


############################################ old fastchat code, needs update to new socket interface.

def send_w_stream(model, messages, temperature=0.1, max_tokens= 500):
    global MODEL_NAME, WORKER_ADDR, CONTROLLER_ADDRESS
    MODEL_NAME = model_nm
    CONTROLLER_ADDRESS = controller_addr
    ret = requests.post(controller_addr + "/refresh_all_workers")
    ret = requests.post(controller_addr + "/list_models")
    models = ret.json()["models"]
    models.sort()

    if model_name not in models:
        print(f'***** Fastchat {model_name} not available')
        print(f'***** Fastchat available models {models}')
        return

    ret = requests.post(
        controller_addr + "/get_worker_address", json={"model": model_nm}
    )
    WORKER_ADDR = ret.json()["address"]
    print(f"worker_addr: {WORKER_ADDR}")
    conv=conv.get_conv_template(model_name)

    for msg in messages:
        role = msg['role']
        if role.lower() == 'user':
            role_index = 0
        else:
            role_index = 1
        conv.append_message(conv.roles[role_index], msg['content'])
    conv.append_message(conv.roles[1], '')
    prompt = conv.get_prompt()

    headers = {"User-Agent": "FastChat Client"}
    gen_params = {
        "model": MODEL_NAME,
        "prompt": prompt,
        "temperature": temperature,
        "max_new_tokens": max_tokens,
        "stop": conv.stop_str,
        "stop_token_ids": conv.stop_token_ids,
        "echo": False,
    }
    response = requests.post(
        WORKER_ADDR + "/worker_generate_stream",
        headers=headers,
        json=gen_params,
        stream=True,
    )

    prev = 0
    for chunk in response.iter_lines(decode_unicode=False, delimiter=b"\0"):
        if chunk:
            data = json.loads(chunk.decode())
            output = data["text"].strip()
            print(output[prev:], end="", flush=True)
            prev = len(output)

    return output



from transformers import AutoModelForCausalLM, AutoTokenizer
import torch
import transformers
from alphawave_pyexts import serverUtils as sv
import inspect


model_name = "tiiuae/falcon-7b-instruct"
print(f"Loading {model_name}")

tokenizer = AutoTokenizer.from_pretrained(model_name, skip_special_tokens=True, spaces_between_special_tokens=False)
pipeline = transformers.pipeline(
    "text-generation",
    model=model_name,
    tokenizer=tokenizer,
    trust_remote_code=True,
    device_map="auto",
)
print(f"server started {model_name} into memory, use falcon_instruct")
sv.server(model=None, tokenizer=tokenizer, pipeline = pipeline, stop_str=['User:'])
#sv.server(model=model, tokenizer=tokenizer, pipeline = None, stop_str=['User:'])



import torch
from transformers import AutoModelForCausalLM, AutoTokenizer
from alphawave_pyexts import serverUtils as sv
from accelerate import load_checkpoint_and_dispatch
import datetime
import os
import requests, socket
from queue import Queue
import json
import sys
import time
import traceback


model_name = "lmsys/vicuna-7b-v1.3"
print(f"Loading {model_name}")

if __name__ == '__main__':

    tokenizer = AutoTokenizer.from_pretrained(model_name, use_fast=False)
    model = AutoModelForCausalLM.from_pretrained(
        model_name,
        #load_in_8bit=True,
        torch_dtype=torch.float16,
        device_map="auto",
        #trust_remote_code=True
    )
    model.tie_weights()
    print(f'server started {model_name} on port 5004, use vicuna_v1.1')
    sv.server(model=model, tokenizer=tokenizer)



from typing import Dict, Any
from abc import ABC, abstractmethod

class PromptMemory:
    def __init__(self):
        self.memory = {}

    def set(self, property: str, value: str):
        self.memory[property] = value

class PromptFunctions:
    pass

class Tokenizer:
    pass

class CommandSchema:
    def __init__(self, type: str, title: str, description: str, properties: Dict[str, Any], required: List[str], returns: str):
        self.type = type
        self.title = title
        self.description = description
        self.properties = properties
        self.required = required
        self.returns = returns

class SetPropertyCommandInput:
    def __init__(self, property: str, value: str):
        self.property = property
        self.value = value

class SchemaBasedCommand(ABC):
    @abstractmethod
    def __init__(self, schema: CommandSchema, title: str = None, description: str = None):
        self.schema = schema
        self.title = title
        self.description = description

    @abstractmethod
    def execute(self, input: Any, memory: PromptMemory, functions: PromptFunctions, tokenizer: Tokenizer):
        pass

class SetPropertyCommand(SchemaBasedCommand):
    def __init__(self, title: str = None, description: str = None):
        schema = CommandSchema(
            type="object",
            title="setProperty",
            description="writes a string value to short term memory",
            properties={
                "property": {
                    "type": "string",
                    "description": "name of the property to set"
                },
                "value": {
                    "type": "string",
                    "description": "value to remember"
                }
            },
            required=["property", "value"],
            returns="confirmation of the assignment"
        )
        super().__init__(schema, title, description)

    def execute(self, input: SetPropertyCommandInput, memory: PromptMemory, functions: PromptFunctions, tokenizer: Tokenizer):
        memory.set(input.property, input.value)
        return f'the "{input.property}" property was updated'



import torch
import transformers
from transformers import AutoModelForCausalLM, AutoTokenizer
import bitsandbytes
from alphawave_pyexts import serverUtils as sv
from accelerate import load_checkpoint_and_dispatch
import datetime
import os
import requests, socket
from queue import Queue
import json
import sys
import time
import traceback


model_name = "/home/bruce/Downloads/llama/llama-2-13b-chat"
print(f'**** Loading {model_name}')

tokenizer = AutoTokenizer.from_pretrained(model_name, use_fast=False, skip_special_tokens=True)
model = AutoModelForCausalLM.from_pretrained(
    model_name,
    torch_dtype=torch.float16,
    #load_in_8bit=True,
    use_safetensors=True,
    device_map="auto",
)
model.tie_weights()
pipeline = transformers.pipeline(
    "text-generation",
    model=model,
    tokenizer=tokenizer,
    device_map="auto",
)

print(f'server started {model_name} on port 5004, use aplaca')
#sv.server(model=model, tokenizer=tokenizer, stop_str=['###', '### Input', '### Response'])
sv.server(tokenizer=tokenizer, pipeline=pipeline, stop_str=['### Instruction', '### User'])



import json
import sys
import os
import time
import traceback
import readline as rl
from alphawave.MemoryFork import MemoryFork
import alphawave_pyexts.llmsearch.google_search_concurrent as gs
import alphawave_pyexts.utilityV2 as ut
import asyncio
history = {}


def run_chat(client, query_string, model,  memory, functions, tokenizer, max_chars=1500, search_level=gs.QUICK_SEARCH):
  #tracemalloc.start()
  response_text = ''
  storeInteraction = True
  try:
    #
    fork = MemoryFork(memory)
    query_phrase, keywords = ut.get_search_phrase_and_keywords(client, query_string, model, fork, functions, tokenizer)

    google_text=\
      gs.search_google(query_string, gs.QUICK_SEARCH, query_phrase, keywords, client, model, fork, functions, tokenizer, max_chars)
    return google_text
  except:
    traceback.print_exc()
  return ''

if __name__ == '__main__' :
  while True:
    run_chat(input('Yes?'))



from transformers import AutoModelForCausalLM, AutoTokenizer, LlamaTokenizer, StoppingCriteria, StoppingCriteriaList, TextIteratorStreamer
import accelerate
import torch
from accelerate import load_checkpoint_and_dispatch
import datetime
import os
from threading import Event, Thread
from uuid import uuid4
import requests, socket
from queue import Queue
import json
import sys
import gc
import time
import string
import traceback
from collections.abc import Iterable

# Load the model.

print('devices', torch.cuda.device_count(), torch.cuda.current_device())
max_new_tokens = 1536

class StopOnTokens(StoppingCriteria):
    def __init__(self, tokenizer, prompt, stop_strs=[]):
      self.stop_ids = []
      self.stop_strs = stop_strs
      self.tokenizer=tokenizer
      self.prompt_len=len(prompt)
      self.prompt=True
      
      if stop_strs is not None and isinstance(stop_strs, Iterable):
        for stop_wd in stop_strs:
          #print(f'***** stop wd pre encode {stop_wd}')
          if stop_wd is not None:
            tokens = tokenizer.encode(stop_wd)
            self.stop_ids.append(tokens)
      print(f'***** stop_id tokens {self.prompt_len}, {self.stop_ids} ')
      
    def __call__(self, input_ids: torch.LongTensor, scores: torch.FloatTensor, **kwargs) -> bool:
      # don't process prompt
      if self.prompt:
        self.prompt = False
        return False
      
      for stop_id in self.stop_ids:
        if len(input_ids[0]) >= len(stop_id):
          if input_ids[0][:-len(stop_id)] == stop_id:
            print(f'***** stopping on token {stop_id}')
            return True
      input_str = self.tokenizer.decode(input_ids[0], skip_special_tokens=True)
      for stop_str in self.stop_strs:
        if len(input_str)>self.prompt_len and stop_str in input_str[self.prompt_len:]:
          print(f'***** stopping on string {stop_str}, \n{input_str[self.prompt_len:]}')
          return True
      return False
    
class MyStreamer(TextIteratorStreamer):
  def __init__(self, tokenizer, stop_event=None, skip_prompt=True):
    super(TextIteratorStreamer, self).__init__(tokenizer, skip_prompt)
    self.stop_event = stop_event
    ### why do these have to be here? should be inherited from TextIteratorStreamer?
    self.text_queue = Queue()
    self.stop_signal = None
    self.timeout = None

  def put(self, value):
    """
        Receives tokens, decodes them, and prints them to stdout as soon as they form entire words.
    """

    if len(value.shape) > 1 and value.shape[0] > 1:
      raise ValueError("TextStreamer only supports batch size 1")
    elif len(value.shape) > 1:
      value = value[0]

    if self.skip_prompt and self.next_tokens_are_prompt:
      self.next_tokens_are_prompt = False
      return

    # Add the new token to the cache and decodes the entire thing.
    self.token_cache.extend(value.tolist())
    text = self.tokenizer.decode(self.token_cache, skip_special_tokens=True)
    
    # After the symbol for a new line, we flush the cache.
    if text.endswith("\n"):
      printable_text = text[self.print_len :]
      self.token_cache = []
      self.print_len = 0
    # If the last token is a CJK character, we print the characters.
    elif len(text) > 0 and self._is_chinese_char(ord(text[-1])):
      printable_text = text[self.print_len :]
      self.print_len += len(printable_text)
      # Otherwise, prints until the last space char (simple heuristic to avoid printing incomplete words,
      # which may change with the subsequent token -- there are probably smarter ways to do this!)
    else:
      printable_text = text[self.print_len : text.rfind(" ") + 1]
      self.print_len += len(printable_text)
    self.on_finalized_text(printable_text)
    if self.stop_event.is_set():
        self.on_finalized_text(" ", stream_end=True)
        print(f"\n***** MyStreamer exit stop_event is {self.stop_event.is_set()}")
        time.sleep(0.1)
        raise SystemExit

  def end(self):
    """Flushes any remaining cache and prints a newline to stdout."""
    # Flush the cache, if it exists
    print('received end from generate')
    if len(self.token_cache) > 0:
      text = self.tokenizer.decode(self.token_cache, skip_special_tokens=True)
      printable_text = text[self.print_len :]
      self.token_cache = []
      self.print_len = 0
    else:
      printable_text = ""
    self.next_tokens_are_prompt = True
    self.on_finalized_text(printable_text, stream_end=True)

  def on_finalized_text(self, text: str, stream_end: bool = False):
      """Put the new text in the queue. If the stream is ending, also put a stop signal in the queue."""
      filtered_text = "".join(filter(lambda x: x in string.printable, text))
      self.text_queue.put(filtered_text, timeout=self.timeout)
      if stream_end:
        self.text_queue.put(self.stop_signal, timeout=self.timeout)

  def __iter__(self):
    return self

  def __next__(self):
    value = self.text_queue.get(timeout=self.timeout)
    if value == self.stop_signal:
      raise StopIteration()
    else:
      return value
    
def submit(message, model=None, tokenizer=None, pipeline=None, stop_event=None,conn=None, stop_str = None):
    stop_event.clear()
    message_j = json.loads(message)
    streamer = MyStreamer(tokenizer, stop_event=stop_event)
    stopping_criteria=StoppingCriteriaList([StopOnTokens(tokenizer, message_j['prompt'], stop_str )])
    temp = 0.7
    if 'temp' in message_j.keys():
        temp = message_j['temp']
    top_p = 1.0
    if 'top_p' in message_j.keys():
        top_p = message_j['top_p']
    max_tokens = 100
    if 'max_tokens' in message_j.keys():
        max_tokens = message_j['max_tokens']
    user_prompt = 'User'
    if 'user' in message_j.keys():
        user_prompt = message_j['user']
    asst_prompt = 'Assistant'
    if 'asst' in message_j.keys():
        asst_prompt = message_j['asst']
    eos = '<|endoftext|>'
    if 'eos' in message_j.keys():
        eos = message_j['eos']

    print(f'\n temp {temp}, top_p: {top_p}, max_tokens={max_tokens}\n')

    ### Pipeline
    if pipeline:
      result = pipeline(message_j['prompt'],
                        max_new_tokens=max_tokens,
                        top_p=top_p,
                        temperature=temp,
                        #do_sample=True,
                        top_k=10,
                        num_return_sequences=1,
                        return_full_text=False,
                        stopping_criteria = stopping_criteria,
                        #eos_token_id=tokenizer.eos_token_id
                        )
      if result is not None and type(result[0]) == dict and 'generated_text' in result[0]:
        generated_text = result[0]['generated_text'] 
        #print(generated_text)
        idx=100000; idx3 = -1; 
        if stop_str is not None and isinstance(stop_str, Iterable):
          for stop in stop_str:
            idx3 = generated_text.find(stop)
            if idx3 >0 and idx3 < idx : idx = idx3
        if idx > 0 and idx < 100000:
          print(f'Truncating at {idx}')
          generated_text = generated_text[:idx]
        conn.send(bytearray((generated_text).encode('utf8')))
        return generated_text
      else: return ''

    ### model and tokenizer
    else:
      encode_dict = tokenizer.encode_plus(message_j['prompt'], return_tensors="pt")# .to(DEV)
      input_ids = encode_dict['input_ids']
      attention_mask = encode_dict['attention_mask']
      input_ids = input_ids.to('cuda')
      generation_kwargs = dict(
        do_sample=True,
        max_new_tokens=max_tokens,
        top_p=top_p,
        temperature=temp,
        streamer = streamer,
        top_k=20,
        num_return_sequences=1,

        attention_mask= attention_mask,
        eos_token_id=tokenizer.eos_token_id
      )
      
      generated_text = ""
      #thread = Thread(target=model.generate, args=(input_ids,), kwargs=generation_kwargs)
      thread = Thread(target=lambda: model.generate(input_ids,
                                                    do_sample=True,
                                                    max_new_tokens=max_tokens,
                                                    min_new_tokens=20,
                                                    top_p=top_p,
                                                    stopping_criteria = stopping_criteria,
                                                    temperature=temp,
                                                    streamer=streamer,
                                                    num_return_sequences=1,
                                                    attention_mask=attention_mask))
      thread.start()
      for new_text in streamer:
        print(new_text)
        test_text = generated_text + new_text
        idx1 = test_text.find('<|endoftext|>')
        idx2 = test_text.find(user_prompt)
        idx3 = -1
        if stop_str is not None and isinstance(stop_str, Iterable):
          for stop in stop_str:
            idx3 = max(idx3, test_text.find(stop))
        if stop_event.is_set(): # flush generate
          continue
        if idx1>= 0 or idx2>=0 or idx3 >= 0:
          stop_event.set()
          # get the earliest stop found
          idx = 1000000
          for candidate_idx in [idx1, idx2, idx3]:
              if candidate_idx >= 0 and candidate_idx < idx:
                  idx = candidate_idx
          print(f'set stop event in submit {idx1}, {idx2}, {idx3}, stop on {idx}')
          if idx < len(test_text) and idx > len(generated_text):  #see if there is anything to send
            conn.send(bytearray((test_text[len(generated_text):idx]).encode('utf8')))
        else:
          generated_text = test_text
          conn.send(bytearray((new_text).encode('utf8')))  # send data to the client
      return generated_text

    
def server(model=None, tokenizer=None, pipeline=None, stop_str=[]):
    global stopping_criteria
    # either model/tokenizer OR pipeline.
    # stop_str is a list of stop_strings
    # get the hostname
    host = socket.gethostname()
    host = ''
    port = 5004  # initiate port no above 1024
    stop_event = Event()
    print(f"starting server")
    

    while True:
      try:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM)  as server_socket: # get instance
            server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            server_socket.bind((host, port))  # bind host address and port together
            server_socket.settimeout(3)

            while True:
                try:
                    conn = None
                    server_socket.listen(30)
                    #print("waiting...")
                    try:
                      conn, address = server_socket.accept()  # accept new connection
                    except Exception:
                        pass
                    if conn is None: continue
                    print("Connection from: " + str(address))
                    query_string = ''
                    while True:
                      s = conn.recv(1024)
                      if s is None or not s:
                        break
                      if (len(s) > 5 and s[-3:] == b'xff' and s[-6:-3] == b'x00'):
                        s = s[:-6]
                        query_string += s.decode('utf-8')
                        break
                      else:
                        query_string += s.decode('utf-8')
                    print("got string:", query_string)
                    if query_string is not None and len(query_string) > 0:
                        message_j = json.loads(query_string)
                        submit(query_string, model=model, tokenizer=tokenizer, pipeline=pipeline, stop_event=stop_event, conn=conn, stop_str=stop_str)
                        gc.collect()
                        torch.cuda.empty_cache()
                        print('sending termination')
                        conn.sendall(b'x00xff')
                        time.sleep(0.5)
                    if conn is not None: conn.close()
                except TimeoutError:
                  #traceback.print_exc()
                    if conn is not None:
                        conn.sendall(b'x00xff')
                        time.sleep(0.5)
                        conn.close()
                except KeyboardInterrupt:
                  print("idle loop interrrupt")
                  traceback.print_exc()
                  sys.exit(0)
        print("resetting socket")
        server_socket.close()
      except BrokenPipeError:
        traceback.print_exc()
        pass
      except KeyboardInterrupt:
        sys.exit(0)
      except Exception:
        traceback.print_exc()




from promptrix.Prompt import Prompt
from promptrix.SystemMessage import SystemMessage
from promptrix.ConversationHistory import ConversationHistory
from promptrix.UserMessage import UserMessage
from promptrix.promptrixTypes import Message
from alphawave.OpenAIClient  import OpenAIClient
from alphawave.AlphaWave import AlphaWave
from alphawave.JSONResponseValidator import JSONResponseValidator
import os
import jsonschema
import readline

# Read in .env file.

# Create an OpenAI or AzureOpenAI client
client = OpenAIClient(api_key=os.getenv("OpenAIKey"))

# Define expected response schema and create a validator
ResponseSchema = {
    "type": "object",
    "properties": {
        "answer": {"type": "string"},
        "sentiment": {"type": "string", "enum": ["positive", "neutral", "negative"]}
    },
    "required": ["answer", "sentiment"]
}

validator = JSONResponseValidator(ResponseSchema)

# Create a wave
wave = AlphaWave(
    client=client,
    prompt=Prompt([
        SystemMessage(
            "Answers the user but also analyze the sentiment of their message.\n"
            "Return your answer using this JSON structure:\n"
            '{"answer":"<answer>","sentiment":"positive|neutral|negative"}'
        ),
        ConversationHistory('history'),
        UserMessage('{{$input}}', 200)
    ]),
    prompt_options={
        "completion_type": "chat",
        "model": "gpt-3.5-turbo",
        "temperature": 0.9,
        "max_input_tokens": 2000,
        "max_tokens": 1000,
    },
    validator=validator,
    logRepairs=True
)

# Define main chat loop
def chat(botMessage=None):
    # Show the bots message
    if botMessage:
        if "sentiment" in botMessage:
            print(f"[{botMessage['sentiment']}]")
        print(f"{botMessage['answer']}")

    # Prompt the user for input
    input = input('User: ')
    # Check if the user wants to exit the chat
    if input.lower() == 'exit':
        # Close the readline interface and exit the process
        exit()
    else:
        # Route users message to wave
        result = wave.completePrompt(input)
        if result.status == 'success':
            chat(result.message.content)
        else:
            if result.message:
                print(f"{result.status}: {result.message}")
            else:
                print(f"A result status of '{result.status}' was returned.")
            exit()

# Start chat session
chat({"answer": "Hello, how can I help you?"})



# port of models described in RW
# We use the bloom model as a starting point for these model.
# Please refer to the bloom models for usage instructions.

import math
import warnings
from typing import Optional, Tuple, Union

import torch
import torch.utils.checkpoint
from torch import nn
from torch.nn import BCEWithLogitsLoss, CrossEntropyLoss, LayerNorm, MSELoss
from torch.nn import functional as F

from transformers.modeling_outputs import (
    BaseModelOutputWithPastAndCrossAttentions,
    CausalLMOutputWithCrossAttentions,
    QuestionAnsweringModelOutput,
    SequenceClassifierOutputWithPast,
    TokenClassifierOutput,
)
from transformers.modeling_utils import PreTrainedModel
from transformers.utils import logging
from .configuration_RW import RWConfig

logger = logging.get_logger(__name__)

# NOTE(Hesslow): Unfortunately we did not fuse matmul and bias during training, this means that there's one additional quantization to bfloat16 between the operations.
# In order not to degrade the quality of our HF-port, we keep these characteristics in the final model.
class Linear(nn.Linear):
    def forward(self, input: torch.Tensor) -> torch.Tensor:
        ret = input @ self.weight.T
        if self.bias is None:
            return ret
        else:
            return ret + self.bias


from einops import rearrange

# rotary pos emb helpers (torch.jit.script does not seem to support staticmethod...)
def rotate_half(x):
    x1, x2 = x[..., : x.shape[-1] // 2], x[..., x.shape[-1] // 2 :]
    return torch.cat((-x2, x1), dim=x1.ndim - 1)  # dim=-1 triggers a bug in torch < 1.8.0


class RotaryEmbedding(torch.nn.Module):
    """Implementation of RotaryEmbedding from GPT-NeoX.
    This implementation is design to operate on queries and keys that are compatible with
    [batch_size, n_heads_per_partition, seq_len, head_dim] (e.g. MinGPTAttention format).
    """

    def __init__(
        self,
        head_dim: int,
        base=10000,
    ):
        super().__init__()
        inv_freq = 1.0 / (base ** (torch.arange(0, head_dim, 2).float() / head_dim))
        self.register_buffer("inv_freq", inv_freq, persistent=False)
        self.head_dim = head_dim
        self.seq_len_cached = None
        self.batch_size_cached = None
        self.cos_cached: torch.Tensor | None = None
        self.sin_cached: torch.Tensor | None = None

    def cos_sin(
        self,
        seq_len: int,
        device="cuda",
        dtype=torch.bfloat16,
    ) -> torch.Tensor:
        if seq_len != self.seq_len_cached:
            self.seq_len_cached = seq_len
            t = torch.arange(seq_len, device=device).type_as(self.inv_freq)
            freqs = torch.einsum("i,j->ij", t, self.inv_freq)
            emb = torch.cat((freqs, freqs), dim=-1).to(device)

            if dtype in [torch.float16, torch.bfloat16]:
                emb = emb.float()

            self.cos_cached = emb.cos()[None, :, :]
            self.sin_cached = emb.sin()[None, :, :]

            self.cos_cached = self.cos_cached.type(dtype)
            self.sin_cached = self.sin_cached.type(dtype)

        return self.cos_cached, self.sin_cached

    def forward(self, q, k):
        batch, seq_len, head_dim = q.shape
        cos, sin = self.cos_sin(seq_len, q.device, q.dtype)
        return (q * cos) + (rotate_half(q) * sin), (k * cos) + (rotate_half(k) * sin)


def _make_causal_mask(
    input_ids_shape: torch.Size, device: torch.device, past_key_values_length: int
) -> torch.BoolTensor:
    batch_size, target_length = input_ids_shape
    mask = torch.empty((target_length, target_length + past_key_values_length), dtype=torch.bool, device=device)
    # ONNX doesn't support `torch.Tensor.triu` properly, thus we use this workaround
    seq_ids = torch.arange(target_length, device=device)
    mask[:, past_key_values_length:] = seq_ids[:, None] < seq_ids[None, :]

    if past_key_values_length > 0:
        mask[:, :past_key_values_length] = False

    expanded_mask = mask[None, None, :, :].expand(batch_size, 1, target_length, target_length + past_key_values_length)
    return expanded_mask


def _expand_mask(mask: torch.Tensor, tgt_length: int) -> torch.BoolTensor:
    batch_size, src_length = mask.shape
    tgt_length = tgt_length if tgt_length is not None else src_length

    expanded_mask = ~(mask[:, None, None, :].to(torch.bool))
    return expanded_mask.expand(batch_size, 1, tgt_length, src_length)


def build_alibi_tensor(attention_mask: torch.Tensor, num_heads: int, dtype: torch.dtype) -> torch.Tensor:
    batch_size, seq_length = attention_mask.shape
    closest_power_of_2 = 2 ** math.floor(math.log2(num_heads))
    base = torch.tensor(
        2 ** (-(2 ** -(math.log2(closest_power_of_2) - 3))), device=attention_mask.device, dtype=torch.float32
    )
    powers = torch.arange(1, 1 + closest_power_of_2, device=attention_mask.device, dtype=torch.int32)
    slopes = torch.pow(base, powers)

    if closest_power_of_2 != num_heads:
        extra_base = torch.tensor(
            2 ** (-(2 ** -(math.log2(2 * closest_power_of_2) - 3))), device=attention_mask.device, dtype=torch.float32
        )
        num_remaining_heads = min(closest_power_of_2, num_heads - closest_power_of_2)
        extra_powers = torch.arange(1, 1 + 2 * num_remaining_heads, 2, device=attention_mask.device, dtype=torch.int32)
        slopes = torch.cat([slopes, torch.pow(extra_base, extra_powers)], dim=0)

    # Note: alibi will added to the attention bias that will be applied to the query, key product of attention
    # => therefore alibi will have to be of shape (batch_size, num_heads, query_length, key_length)
    # => here we set (batch_size=1, num_heads=num_heads, query_length=1, key_length=max_length)
    # => the query_length dimension will then be broadcasted correctly
    # This is more or less identical to T5's relative position bias:
    # https://github.com/huggingface/transformers/blob/f681437203baa7671de3174b0fa583c349d9d5e1/src/transformers/models/t5/modeling_t5.py#L527
    arange_tensor = ((attention_mask.cumsum(dim=-1) - 1) * attention_mask)[:, None, :]
    alibi = slopes[..., None].bfloat16() * arange_tensor
    return alibi.reshape(batch_size * num_heads, 1, seq_length).to(dtype)


def dropout_add(x: torch.Tensor, residual: torch.Tensor, prob: float, training: bool) -> torch.Tensor:
    out = F.dropout(x, p=prob, training=training)
    out = residual + out
    return out


class Attention(nn.Module):
    def __init__(self, config: RWConfig):
        super().__init__()

        self.hidden_size = config.hidden_size
        self.num_heads = config.n_head
        self.head_dim = self.hidden_size // self.num_heads
        self.split_size = self.hidden_size
        self.hidden_dropout = config.hidden_dropout

        if self.head_dim * self.num_heads != self.hidden_size:
            raise ValueError(
                f"`hidden_size` must be divisible by num_heads (got `hidden_size`: {self.hidden_size} and `num_heads`:"
                f" {self.num_heads})."
            )

        self.maybe_rotary = RotaryEmbedding(config.head_dim) if config.rotary else lambda q, k: (q, k)

        # Layer-wise attention scaling
        self.inv_norm_factor = 1.0 / math.sqrt(self.head_dim)
        self.beta = self.inv_norm_factor

        self.query_key_value = Linear(
            self.hidden_size,
            3 * self.hidden_size if not config.multi_query else (self.hidden_size + 2 * self.head_dim),
            bias=config.bias,
        )
        self.multi_query = config.multi_query
        self.dense = Linear(self.hidden_size, self.hidden_size, bias=config.bias)
        self.attention_dropout = nn.Dropout(config.attention_dropout)
        self.num_kv = config.n_head if not self.multi_query else 1

    def _split_heads(self, fused_qkv: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Split the last dimension into (num_heads, head_dim) without making any copies, results share same memory
        storage as `fused_qkv`

        Args:
            fused_qkv (`torch.tensor`, *required*): [batch_size, seq_length, num_heads * 3 * head_dim]

        Returns:
            query: [batch_size, seq_length, num_heads, head_dim] key: [batch_size, seq_length, num_heads, head_dim]
            value: [batch_size, seq_length, num_heads, head_dim]
        """
        if not self.multi_query:
            batch_size, seq_length, three_times_hidden_size = fused_qkv.shape
            fused_qkv = fused_qkv.view(batch_size, seq_length, self.num_heads, 3, self.head_dim)
            return fused_qkv[..., 0, :], fused_qkv[..., 1, :], fused_qkv[..., 2, :]
        else:
            batch_size, seq_length, three_times_hidden_size = fused_qkv.shape
            fused_qkv = fused_qkv.view(batch_size, seq_length, self.num_heads + 2, self.head_dim)
            return fused_qkv[..., :-2, :], fused_qkv[..., [-2], :], fused_qkv[..., [-1], :]

    def _merge_heads(self, x: torch.Tensor) -> torch.Tensor:
        """
        Merge heads together over the last dimenstion

        Args:
            x: (`torch.tensor`, *required*): [batch_size * num_heads, seq_length, head_dim]

        Returns:
            torch.tensor: [batch_size, seq_length, num_heads * head_dim]
        """
        # What we want to achieve is:
        # batch_size * num_heads, seq_length, head_dim -> batch_size, seq_length, num_heads * head_dim
        batch_size_and_num_heads, seq_length, _ = x.shape
        batch_size = batch_size_and_num_heads // self.num_heads

        # First view to decompose the batch size
        # batch_size * num_heads, seq_length, head_dim -> batch_size, num_heads, seq_length, head_dim
        x = x.view(batch_size, self.num_heads, seq_length, self.head_dim)

        # batch_size, num_heads, seq_length, head_dim -> batch_size, seq_length, num_heads, head_dim
        x = x.permute(0, 2, 1, 3)

        # batch_size, seq_length, num_heads, head_dim -> batch_size, seq_length, num_heads * head_dim
        return x.reshape(batch_size, seq_length, self.num_heads * self.head_dim)

    def forward(
        self,
        hidden_states: torch.Tensor,
        alibi: torch.Tensor,
        attention_mask: torch.Tensor,
        layer_past: Optional[Tuple[torch.Tensor, torch.Tensor]] = None,
        head_mask: Optional[torch.Tensor] = None,
        use_cache: bool = False,
        output_attentions: bool = False,
    ):
        fused_qkv = self.query_key_value(hidden_states)  # [batch_size, seq_length, 3 x hidden_size]

        # 3 x [batch_size, seq_length, num_heads, head_dim]
        (query_layer, key_layer, value_layer) = self._split_heads(fused_qkv)

        batch_size, q_length, _, _ = query_layer.shape

        query_layer = query_layer.transpose(1, 2).reshape(batch_size * self.num_heads, q_length, self.head_dim)
        key_layer = key_layer.transpose(1, 2).reshape(
            batch_size * self.num_kv,
            q_length,
            self.head_dim,
        )
        value_layer = value_layer.transpose(1, 2).reshape(batch_size * self.num_kv, q_length, self.head_dim)

        query_layer, key_layer = self.maybe_rotary(query_layer, key_layer)

        if layer_past is not None:
            past_key, past_value = layer_past
            # concatenate along seq_length dimension:
            #  - key: [batch_size * self.num_heads, head_dim, kv_length]
            #  - value: [batch_size * self.num_heads, kv_length, head_dim]
            key_layer = torch.cat((past_key, key_layer), dim=1)
            value_layer = torch.cat((past_value, value_layer), dim=1)

        _, kv_length, _ = key_layer.shape

        if use_cache is True:
            present = (key_layer, value_layer)
        else:
            present = None

        if alibi is None:
            query_layer_ = query_layer.reshape(batch_size, self.num_heads, -1, self.head_dim)
            key_layer_ = key_layer.reshape(batch_size, self.num_kv, -1, self.head_dim)
            value_layer_ = value_layer.reshape(batch_size, self.num_kv, -1, self.head_dim)

            attn_output = F.scaled_dot_product_attention(
                query_layer_, key_layer_, value_layer_, None, 0.0, is_causal=True
            )

            x = attn_output.view(batch_size, self.num_heads, q_length, self.head_dim)
            x = x.permute(0, 2, 1, 3)
            attn_output = x.reshape(batch_size, q_length, self.num_heads * self.head_dim)

            output_tensor = self.dense(attn_output)

            outputs = (output_tensor, present)
            assert not output_attentions  # not supported.
            return outputs
        else:
            attention_mask_float = (attention_mask * 1.0).masked_fill(attention_mask, -1e9).to(torch.bfloat16)
            matmul_result = query_layer @ key_layer.transpose(-1, -2)

            # change view to [batch_size, num_heads, q_length, kv_length]
            attention_scores = matmul_result.view(batch_size, self.num_heads, q_length, kv_length)

            # cast attention scores to fp32, compute scaled softmax and cast back to initial dtype - [batch_size, num_heads, q_length, kv_length]
            input_dtype = attention_scores.dtype
            # `float16` has a minimum value of -65504.0, whereas `bfloat16` and `float32` have a minimum value of `-3.4e+38`
            if input_dtype == torch.float16 or input_dtype == torch.bfloat16:
                attention_scores = attention_scores.to(torch.float32)
            # attn_weights = torch.masked_fill(attention_scores, attention_mask, torch.finfo(attention_scores.dtype).min)
            attention_probs = F.softmax(
                (attention_scores + alibi) * self.inv_norm_factor + attention_mask_float,
                dim=-1,
                dtype=hidden_states.dtype,
            )
            # [batch_size, num_heads, q_length, kv_length]
            attention_probs = self.attention_dropout(attention_probs)

            if head_mask is not None:
                attention_probs = attention_probs * head_mask

            # change view [batch_size x num_heads, q_length, kv_length]
            attention_probs_reshaped = attention_probs.view(batch_size * self.num_heads, q_length, kv_length)

            # matmul: [batch_size * num_heads, q_length, head_dim]
            context_layer = attention_probs_reshaped @ value_layer

            # change view [batch_size, num_heads, q_length, head_dim]
            context_layer = self._merge_heads(context_layer)

            output_tensor = self.dense(context_layer)

            outputs = (output_tensor, present)
            if output_attentions:
                outputs += (attention_probs,)

            return outputs


class MLP(nn.Module):
    def __init__(self, config: RWConfig):
        super().__init__()
        hidden_size = config.hidden_size

        self.dense_h_to_4h = Linear(hidden_size, 4 * hidden_size, bias=config.bias)
        self.act = nn.GELU()
        self.dense_4h_to_h = Linear(4 * hidden_size, hidden_size, bias=config.bias)
        self.hidden_dropout = config.hidden_dropout

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.act(self.dense_h_to_4h(x))
        x = self.dense_4h_to_h(x)
        return x


class DecoderLayer(nn.Module):
    def __init__(self, config: RWConfig):
        super().__init__()
        hidden_size = config.hidden_size

        self.input_layernorm = LayerNorm(hidden_size, eps=config.layer_norm_epsilon)
        self.num_heads = config.n_head
        self.self_attention = Attention(config)

        if not config.parallel_attn:
            # unused if parallel attn
            self.post_attention_layernorm = LayerNorm(hidden_size, eps=config.layer_norm_epsilon)

        self.mlp = MLP(config)

        self.apply_residual_connection_post_layernorm = config.apply_residual_connection_post_layernorm
        self.hidden_dropout = config.hidden_dropout

        self.config = config

    def forward(
        self,
        hidden_states: torch.Tensor,
        alibi: torch.Tensor,
        attention_mask: torch.Tensor,
        layer_past: Optional[Tuple[torch.Tensor, torch.Tensor]] = None,
        head_mask: Optional[torch.Tensor] = None,
        use_cache: bool = False,
        output_attentions: bool = False,
    ):

        layernorm_output = self.input_layernorm(hidden_states)
        residual = hidden_states

        # Self attention.
        attn_outputs = self.self_attention(
            layernorm_output,
            layer_past=layer_past,
            attention_mask=attention_mask,
            alibi=alibi,
            head_mask=head_mask,
            use_cache=use_cache,
            output_attentions=output_attentions,
        )

        attention_output = attn_outputs[0]

        if not self.config.parallel_attn:
            residual = dropout_add(attention_output, residual, self.config.attention_dropout, training=self.training)
            layernorm_output = self.post_attention_layernorm(residual)

        outputs = attn_outputs[1:]

        # MLP.
        mlp_output = self.mlp(layernorm_output)

        if self.config.parallel_attn:
            mlp_output += attention_output

        output = dropout_add(mlp_output, residual, self.config.hidden_dropout, training=self.training)

        if use_cache:
            outputs = (output,) + outputs
        else:
            outputs = (output,) + outputs[1:]

        return outputs  # hidden_states, present, attentions


class RWPreTrainedModel(PreTrainedModel):
    _keys_to_ignore_on_load_missing = [r"h.*.self_attention.scale_mask_softmax.causal_mask", r"lm_head.weight"]
    """
    An abstract class to handle weights initialization and a simple interface for downloading and loading pretrained
    models.
    """

    config_class = RWConfig
    base_model_prefix = "transformer"
    supports_gradient_checkpointing = True
    _no_split_modules = ["DecoderLayer"]

    def __init__(self, *inputs, **kwargs):
        super().__init__(*inputs, **kwargs)

    def _init_weights(self, module: nn.Module):
        """Initialize the weights."""
        if isinstance(module, nn.Linear) or isinstance(module, Linear):
            # Slightly different from the TF version which uses truncated_normal for initialization
            # cf https://github.com/pytorch/pytorch/pull/5617
            module.weight.data.normal_(mean=0.0, std=self.config.initializer_range)
            if module.bias is not None:
                module.bias.data.zero_()
        elif isinstance(module, nn.Embedding):
            module.weight.data.normal_(mean=0.0, std=self.config.initializer_range)
            if module.padding_idx is not None:
                module.weight.data[module.padding_idx].zero_()
        elif isinstance(module, LayerNorm):
            module.bias.data.zero_()
            module.weight.data.fill_(1.0)

    def _set_gradient_checkpointing(self, module: nn.Module, value: bool = False):
        if isinstance(module, RWModel):
            module.gradient_checkpointing = value

    @staticmethod
    def _convert_to_standard_cache(
        past_key_value: Tuple[Tuple[torch.Tensor, torch.Tensor]], batch_size: int
    ) -> Tuple[Tuple[torch.Tensor, torch.Tensor]]:
        """
        Standardizes the format of the cache so as to match most implementations, i.e. to tuple(tuple([batch_size,
        num_heads, ...]))
        """
        batch_size_times_num_heads, head_dim, seq_length = past_key_value[0][0].shape
        num_heads = batch_size_times_num_heads // batch_size
        # key: [batch_size * num_heads, head_dim, seq_length] -> [batch_size, num_heads, head_dim, seq_length]
        # value: [batch_size * num_heads, seq_length, head_dim] -> [batch_size, num_heads, seq_length, head_dim]
        return tuple(
            (
                layer_past[0].view(batch_size, num_heads, head_dim, seq_length),
                layer_past[1].view(batch_size, num_heads, seq_length, head_dim),
            )
            for layer_past in past_key_value
        )

    @staticmethod
    def _convert_to_rw_cache(
        past_key_value: Tuple[Tuple[torch.Tensor, torch.Tensor]]
    ) -> Tuple[Tuple[torch.Tensor, torch.Tensor]]:
        batch_size, num_heads, head_dim, seq_length = past_key_value[0][0].shape
        batch_size_times_num_heads = batch_size * num_heads
        # key:  [batch_size, num_heads, head_dim, seq_length] -> [batch_size * num_heads, head_dim, seq_length]
        # value: [batch_size, num_heads, seq_length, head_dim] -> [batch_size * num_heads, seq_length, head_dim]
        return tuple(
            (
                layer_past[0].view(batch_size_times_num_heads, head_dim, seq_length),
                layer_past[1].view(batch_size_times_num_heads, seq_length, head_dim),
            )
            for layer_past in past_key_value
        )


class RWModel(RWPreTrainedModel):
    def __init__(self, config: RWConfig):
        super().__init__(config)

        self.embed_dim = config.hidden_size
        self.num_heads = config.n_head
        self.alibi = config.alibi

        # Embedding + LN Embedding
        self.word_embeddings = nn.Embedding(config.vocab_size, self.embed_dim)

        # Transformer blocks
        self.h = nn.ModuleList([DecoderLayer(config) for _ in range(config.num_hidden_layers)])

        # Final Layer Norm
        self.ln_f = LayerNorm(self.embed_dim, eps=config.layer_norm_epsilon)

        self.gradient_checkpointing = False

        # Initialize weights and apply final processing
        self.post_init()

    def get_input_embeddings(self):
        return self.word_embeddings

    def _prepare_attn_mask(
        self, attention_mask: torch.Tensor, input_shape: Tuple[int, int], past_key_values_length: int
    ) -> torch.BoolTensor:
        # create causal mask
        # [batch_size, seq_length] -> [batch_size, 1, tgt_length, src_length]
        combined_attention_mask = None
        device = attention_mask.device
        _, src_length = input_shape

        if src_length > 1:
            combined_attention_mask = _make_causal_mask(
                input_shape, device=device, past_key_values_length=past_key_values_length
            )

        # [batch_size, seq_length] -> [batch_size, 1, tgt_length, src_length]
        expanded_attn_mask = _expand_mask(attention_mask, tgt_length=src_length)
        combined_attention_mask = (
            expanded_attn_mask if combined_attention_mask is None else expanded_attn_mask | combined_attention_mask
        )

        return combined_attention_mask

    def set_input_embeddings(self, new_embeddings: torch.Tensor):
        self.word_embeddings = new_embeddings

    def forward(
        self,
        input_ids: Optional[torch.LongTensor] = None,
        past_key_values: Optional[Tuple[Tuple[torch.Tensor, torch.Tensor], ...]] = None,
        attention_mask: Optional[torch.Tensor] = None,
        head_mask: Optional[torch.LongTensor] = None,
        inputs_embeds: Optional[torch.LongTensor] = None,
        use_cache: Optional[bool] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
        **deprecated_arguments,
    ) -> Union[Tuple[torch.Tensor, ...], BaseModelOutputWithPastAndCrossAttentions]:
        if deprecated_arguments.pop("position_ids", False) is not False:
            # `position_ids` could have been `torch.Tensor` or `None` so defaulting pop to `False` allows to detect if users were passing explicitly `None`
            warnings.warn(
                "`position_ids` have no functionality in BLOOM and will be removed in v5.0.0. You can safely ignore"
                " passing `position_ids`.",
                FutureWarning,
            )
        if len(deprecated_arguments) > 0:
            raise ValueError(f"Got unexpected arguments: {deprecated_arguments}")

        output_attentions = output_attentions if output_attentions is not None else self.config.output_attentions
        output_hidden_states = (
            output_hidden_states if output_hidden_states is not None else self.config.output_hidden_states
        )
        use_cache = use_cache if use_cache is not None else self.config.use_cache
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        if input_ids is not None and inputs_embeds is not None:
            raise ValueError("You cannot specify both input_ids and inputs_embeds at the same time")
        elif input_ids is not None:
            batch_size, seq_length = input_ids.shape
        elif inputs_embeds is not None:
            batch_size, seq_length, _ = inputs_embeds.shape
        else:
            raise ValueError("You have to specify either input_ids or inputs_embeds")

        if past_key_values is None:
            past_key_values = tuple([None] * len(self.h))

        # Prepare head mask if needed
        # 1.0 in head_mask indicate we keep the head
        # attention_probs has shape batch_size x num_heads x N x N
        # head_mask has shape n_layer x batch x num_heads x N x N
        head_mask = self.get_head_mask(head_mask, self.config.n_layer)

        if inputs_embeds is None:
            inputs_embeds = self.word_embeddings(input_ids)

        hidden_states = inputs_embeds

        presents = () if use_cache else None
        all_self_attentions = () if output_attentions else None
        all_hidden_states = () if output_hidden_states else None

        # Compute alibi tensor: check build_alibi_tensor documentation
        seq_length_with_past = seq_length
        past_key_values_length = 0
        if past_key_values[0] is not None:
            past_key_values_length = past_key_values[0][0].shape[2]
            seq_length_with_past = seq_length_with_past + past_key_values_length
        if attention_mask is None:
            attention_mask = torch.ones((batch_size, seq_length_with_past), device=hidden_states.device)
        else:
            attention_mask = attention_mask.to(hidden_states.device)

        if self.alibi:
            alibi = build_alibi_tensor(attention_mask, self.num_heads, dtype=hidden_states.dtype)
        else:
            alibi = None

        causal_mask = self._prepare_attn_mask(
            attention_mask,
            input_shape=(batch_size, seq_length),
            past_key_values_length=past_key_values_length,
        )

        for i, (block, layer_past) in enumerate(zip(self.h, past_key_values)):

            if output_hidden_states:
                all_hidden_states = all_hidden_states + (hidden_states,)

            if self.gradient_checkpointing and self.training:

                if use_cache:
                    logger.warning(
                        "`use_cache=True` is incompatible with gradient checkpointing. Setting `use_cache=False`..."
                    )
                    use_cache = False

                def create_custom_forward(module):
                    def custom_forward(*inputs):
                        # None for past_key_value
                        return module(*inputs, use_cache=use_cache, output_attentions=output_attentions)

                    return custom_forward

                outputs = torch.utils.checkpoint.checkpoint(
                    create_custom_forward(block),
                    hidden_states,
                    alibi,
                    causal_mask,
                    head_mask[i],
                )
            else:
                outputs = block(
                    hidden_states,
                    layer_past=layer_past,
                    attention_mask=causal_mask,
                    head_mask=head_mask[i],
                    use_cache=use_cache,
                    output_attentions=output_attentions,
                    alibi=alibi,
                )

            hidden_states = outputs[0]
            if use_cache is True:
                presents = presents + (outputs[1],)

            if output_attentions:
                all_self_attentions = all_self_attentions + (outputs[2 if use_cache else 1],)

        # Add last hidden state
        hidden_states = self.ln_f(hidden_states)

        if output_hidden_states:
            all_hidden_states = all_hidden_states + (hidden_states,)

        if not return_dict:
            return tuple(v for v in [hidden_states, presents, all_hidden_states, all_self_attentions] if v is not None)

        return BaseModelOutputWithPastAndCrossAttentions(
            last_hidden_state=hidden_states,
            past_key_values=presents,
            hidden_states=all_hidden_states,
            attentions=all_self_attentions,
        )


class RWForCausalLM(RWPreTrainedModel):
    _keys_to_ignore_on_load_missing = [r"h.*.self_attention.scale_mask_softmax.causal_mask", r"lm_head.weight"]

    def __init__(self, config: RWConfig):
        super().__init__(config)
        self.transformer = RWModel(config)
        self.lm_head = nn.Linear(config.hidden_size, config.vocab_size, bias=False)

        # Initialize weights and apply final processing
        self.post_init()

    def get_output_embeddings(self):
        return self.lm_head

    def set_output_embeddings(self, new_embeddings: torch.Tensor):
        self.lm_head = new_embeddings

    def prepare_inputs_for_generation(
        self,
        input_ids: torch.LongTensor,
        past: Optional[torch.Tensor] = None,
        attention_mask: Optional[torch.Tensor] = None,
        **kwargs,
    ) -> dict:
        # only last token for input_ids if past is not None
        if past:
            input_ids = input_ids[:, -1].unsqueeze(-1)

            # the cache may be in the stardard format (e.g. in contrastive search), convert to our's format if needed
            if past[0][0].shape[0] == input_ids.shape[0]:
                past = self._convert_to_rw_cache(past)

        return {
            "input_ids": input_ids,
            "past_key_values": past,
            "use_cache": kwargs.get("use_cache"),
            "attention_mask": attention_mask,
        }

    def forward(
        self,
        input_ids: Optional[torch.LongTensor] = None,
        past_key_values: Optional[Tuple[Tuple[torch.Tensor, torch.Tensor], ...]] = None,
        attention_mask: Optional[torch.Tensor] = None,
        head_mask: Optional[torch.Tensor] = None,
        inputs_embeds: Optional[torch.Tensor] = None,
        labels: Optional[torch.Tensor] = None,
        use_cache: Optional[bool] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
        **deprecated_arguments,
    ) -> Union[Tuple[torch.Tensor], CausalLMOutputWithCrossAttentions]:
        r"""
        labels (`torch.LongTensor` of shape `(batch_size, sequence_length)`, *optional*):
            Labels for language modeling. Note that the labels **are shifted** inside the model, i.e. you can set
            `labels = input_ids` Indices are selected in `[-100, 0, ..., config.vocab_size]` All labels set to `-100`
            are ignored (masked), the loss is only computed for labels in `[0, ..., config.vocab_size]`
        """
        if deprecated_arguments.pop("position_ids", False) is not False:
            # `position_ids` could have been `torch.Tensor` or `None` so defaulting pop to `False` allows to detect if users were passing explicitly `None`
            warnings.warn(
                "`position_ids` have no functionality in BLOOM and will be removed in v5.0.0. You can safely ignore"
                " passing `position_ids`.",
                FutureWarning,
            )
        if len(deprecated_arguments) > 0:
            raise ValueError(f"Got unexpected arguments: {deprecated_arguments}")

        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        transformer_outputs = self.transformer(
            input_ids,
            past_key_values=past_key_values,
            attention_mask=attention_mask,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )
        hidden_states = transformer_outputs[0]

        lm_logits = self.lm_head(hidden_states)

        loss = None
        if labels is not None:
            # Shift so that tokens < n predict n
            shift_logits = lm_logits[..., :-1, :].contiguous()
            shift_labels = labels[..., 1:].contiguous()
            batch_size, seq_length, vocab_size = shift_logits.shape
            # Flatten the tokens
            loss_fct = CrossEntropyLoss()
            loss = loss_fct(
                shift_logits.view(batch_size * seq_length, vocab_size), shift_labels.view(batch_size * seq_length)
            )

        if not return_dict:
            output = (lm_logits,) + transformer_outputs[1:]
            return ((loss,) + output) if loss is not None else output

        return CausalLMOutputWithCrossAttentions(
            loss=loss,
            logits=lm_logits,
            past_key_values=transformer_outputs.past_key_values,
            hidden_states=transformer_outputs.hidden_states,
            attentions=transformer_outputs.attentions,
        )

    def _reorder_cache(
        self, past: Tuple[Tuple[torch.Tensor, torch.Tensor], ...], beam_idx: torch.LongTensor
    ) -> Tuple[Tuple[torch.Tensor, torch.Tensor], ...]:
        """
        This function is used to re-order the `past_key_values` cache if [`~PreTrainedModel.beam_search`] or
        [`~PreTrainedModel.beam_sample`] is called. This is required to match `past_key_values` with the correct
        beam_idx at every generation step.

        Output shares the same memory storage as `past`.
        """
        standardized_past = self._convert_to_standard_cache(past, batch_size=len(beam_idx))

        # Get a copy of `beam_idx` on all the devices where we need those indices.
        device_to_beam_idx = {
            past_state.device: beam_idx.to(past_state.device) for layer_past in past for past_state in layer_past
        }
        reordered_past = tuple(
            (
                layer_past[0].index_select(0, device_to_beam_idx[layer_past[0].device]),
                layer_past[1].index_select(0, device_to_beam_idx[layer_past[0].device]),
            )
            for layer_past in standardized_past
        )
        return self._convert_to_rw_cache(reordered_past)


class RWForSequenceClassification(RWPreTrainedModel):
    _keys_to_ignore_on_load_missing = [r"h.*.self_attention.scale_mask_softmax.causal_mask", r"lm_head.weight"]

    def __init__(self, config: RWConfig):
        super().__init__(config)
        self.num_labels = config.num_labels
        self.transformer = RWModel(config)
        self.score = nn.Linear(config.hidden_size, config.num_labels, bias=False)

        # Initialize weights and apply final processing
        self.post_init()

    def forward(
        self,
        input_ids: Optional[torch.LongTensor] = None,
        past_key_values: Optional[Tuple[Tuple[torch.Tensor, torch.Tensor], ...]] = None,
        attention_mask: Optional[torch.Tensor] = None,
        head_mask: Optional[torch.Tensor] = None,
        inputs_embeds: Optional[torch.Tensor] = None,
        labels: Optional[torch.Tensor] = None,
        use_cache: Optional[bool] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
        **deprecated_arguments,
    ) -> Union[Tuple[torch.Tensor], SequenceClassifierOutputWithPast]:
        r"""
        labels (`torch.LongTensor` of shape `(batch_size,)`, *optional*):
            Labels for computing the sequence classification/regression loss. Indices should be in `[0, ...,
            config.num_labels - 1]`. If `config.num_labels == 1` a regression loss is computed (Mean-Square loss), If
            `config.num_labels > 1` a classification loss is computed (Cross-Entropy).
        """
        if deprecated_arguments.pop("position_ids", False) is not False:
            # `position_ids` could have been `torch.Tensor` or `None` so defaulting pop to `False` allows to detect if users were passing explicitly `None`
            warnings.warn(
                "`position_ids` have no functionality in BLOOM and will be removed in v5.0.0. You can safely ignore"
                " passing `position_ids`.",
                FutureWarning,
            )
        if len(deprecated_arguments) > 0:
            raise ValueError(f"Got unexpected arguments: {deprecated_arguments}")

        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        transformer_outputs = self.transformer(
            input_ids,
            past_key_values=past_key_values,
            attention_mask=attention_mask,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )

        hidden_states = transformer_outputs[0]
        logits = self.score(hidden_states)

        if input_ids is not None:
            batch_size = input_ids.shape[0]
        else:
            batch_size = inputs_embeds.shape[0]

        if self.config.pad_token_id is None and batch_size != 1:
            raise ValueError("Cannot handle batch sizes > 1 if no padding token is defined.")
        if self.config.pad_token_id is None:
            sequence_lengths = -1
        else:
            if input_ids is not None:
                sequence_lengths = torch.ne(input_ids, self.config.pad_token_id).sum(dim=-1) - 1
            else:
                sequence_lengths = -1
                logger.warning(
                    f"{self.__class__.__name__} will not detect padding tokens in `inputs_embeds`. Results may be "
                    "unexpected if using padding tokens in conjunction with `inputs_embeds.`"
                )

        pooled_logits = logits[torch.arange(batch_size, device=logits.device), sequence_lengths]

        loss = None
        if labels is not None:
            if self.config.problem_type is None:
                if self.num_labels == 1:
                    self.config.problem_type = "regression"
                elif self.num_labels > 1 and (labels.dtype == torch.long or labels.dtype == torch.int):
                    self.config.problem_type = "single_label_classification"
                else:
                    self.config.problem_type = "multi_label_classification"

            if self.config.problem_type == "regression":
                loss_fct = MSELoss()
                if self.num_labels == 1:
                    loss = loss_fct(pooled_logits.squeeze(), labels.squeeze())
                else:
                    loss = loss_fct(pooled_logits, labels)
            elif self.config.problem_type == "single_label_classification":
                loss_fct = CrossEntropyLoss()
                loss = loss_fct(pooled_logits, labels)
            elif self.config.problem_type == "multi_label_classification":
                loss_fct = BCEWithLogitsLoss()
                loss = loss_fct(pooled_logits, labels)
        if not return_dict:
            output = (pooled_logits,) + transformer_outputs[1:]
            return ((loss,) + output) if loss is not None else output

        return SequenceClassifierOutputWithPast(
            loss=loss,
            logits=pooled_logits,
            past_key_values=transformer_outputs.past_key_values,
            hidden_states=transformer_outputs.hidden_states,
            attentions=transformer_outputs.attentions,
        )


class RWForTokenClassification(RWPreTrainedModel):
    _keys_to_ignore_on_load_missing = [r"h.*.self_attention.scale_mask_softmax.causal_mask", r"lm_head.weight"]

    def __init__(self, config: RWConfig):
        super().__init__(config)
        self.num_labels = config.num_labels

        self.transformer = RWModel(config)
        if hasattr(config, "classifier_dropout") and config.classifier_dropout is not None:
            classifier_dropout = config.classifier_dropout
        elif hasattr(config, "hidden_dropout") and config.hidden_dropout is not None:
            classifier_dropout = config.hidden_dropout
        else:
            classifier_dropout = 0.1
        self.dropout = nn.Dropout(classifier_dropout)
        self.classifier = nn.Linear(config.hidden_size, config.num_labels)

        # Initialize weights and apply final processing
        self.post_init()

    def forward(
        self,
        input_ids: Optional[torch.LongTensor] = None,
        past_key_values: Optional[Tuple[Tuple[torch.Tensor, torch.Tensor], ...]] = None,
        attention_mask: Optional[torch.Tensor] = None,
        head_mask: Optional[torch.Tensor] = None,
        inputs_embeds: Optional[torch.Tensor] = None,
        labels: Optional[torch.Tensor] = None,
        use_cache: Optional[bool] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
        **deprecated_arguments,
    ) -> Union[Tuple[torch.Tensor], TokenClassifierOutput]:
        r"""
        labels (`torch.LongTensor` of shape `(batch_size,)`, *optional*):
            Labels for computing the sequence classification/regression loss. Indices should be in `[0, ...,
            config.num_labels - 1]`. If `config.num_labels == 1` a regression loss is computed (Mean-Square loss), If
            `config.num_labels > 1` a classification loss is computed (Cross-Entropy).
        """
        if deprecated_arguments.pop("position_ids", False) is not False:
            # `position_ids` could have been `torch.Tensor` or `None` so defaulting pop to `False` allows to detect if users were passing explicitly `None`
            warnings.warn(
                "`position_ids` have no functionality in BLOOM and will be removed in v5.0.0. You can safely ignore"
                " passing `position_ids`.",
                FutureWarning,
            )
        if len(deprecated_arguments) > 0:
            raise ValueError(f"Got unexpected arguments: {deprecated_arguments}")

        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        transformer_outputs = self.transformer(
            input_ids,
            past_key_values=past_key_values,
            attention_mask=attention_mask,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )

        hidden_states = transformer_outputs[0]
        hidden_states = self.dropout(hidden_states)
        logits = self.classifier(hidden_states)

        loss = None
        if labels is not None:
            batch_size, seq_length = labels.shape
            loss_fct = CrossEntropyLoss()
            loss = loss_fct(logits.view(batch_size * seq_length, self.num_labels), labels.view(batch_size * seq_length))

        if not return_dict:
            output = (logits,) + transformer_outputs[2:]
            return ((loss,) + output) if loss is not None else output

        return TokenClassifierOutput(
            loss=loss,
            logits=logits,
            hidden_states=transformer_outputs.hidden_states,
            attentions=transformer_outputs.attentions,
        )


class RWForQuestionAnswering(RWPreTrainedModel):
    _keys_to_ignore_on_load_missing = [r"h.*.self_attention.scale_mask_softmax.causal_mask", r"lm_head.weight"]

    def __init__(self, config):
        super().__init__(config)
        self.transformer = RWModel(config)
        self.qa_outputs = nn.Linear(config.hidden_size, 2)

        # Initialize weights and apply final processing
        self.post_init()

    def forward(
        self,
        input_ids: Optional[torch.LongTensor] = None,
        attention_mask: Optional[torch.FloatTensor] = None,
        position_ids: Optional[torch.LongTensor] = None,
        head_mask: Optional[torch.FloatTensor] = None,
        inputs_embeds: Optional[torch.FloatTensor] = None,
        start_positions: Optional[torch.LongTensor] = None,
        end_positions: Optional[torch.LongTensor] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
    ) -> Union[Tuple, QuestionAnsweringModelOutput]:
        r"""
        start_positions (`torch.LongTensor` of shape `(batch_size,)`, *optional*):
            Labels for position (index) of the start of the labelled span for computing the token classification loss.
            Positions are clamped to the length of the sequence (`sequence_length`). Position outside of the sequence
            are not taken into account for computing the loss.
        end_positions (`torch.LongTensor` of shape `(batch_size,)`, *optional*):
            Labels for position (index) of the end of the labelled span for computing the token classification loss.
            Positions are clamped to the length of the sequence (`sequence_length`). Position outside of the sequence
            are not taken into account for computing the loss.
        """
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        outputs = self.transformer(
            input_ids,
            attention_mask=attention_mask,
            position_ids=position_ids,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )

        sequence_output = outputs[0]

        logits = self.qa_outputs(sequence_output)
        start_logits, end_logits = logits.split(1, dim=-1)
        start_logits = start_logits.squeeze(-1).contiguous()
        end_logits = end_logits.squeeze(-1).contiguous()

        total_loss = None
        if start_positions is not None and end_positions is not None:
            # If we are on multi-GPU, split add a dimension
            if len(start_positions.size()) > 1:
                start_positions = start_positions.squeeze(-1)
            if len(end_positions.size()) > 1:
                end_positions = end_positions.squeeze(-1)
            # sometimes the start/end positions are outside our model inputs, we ignore these terms
            ignored_index = start_logits.size(1)
            start_positions = start_positions.clamp(0, ignored_index)
            end_positions = end_positions.clamp(0, ignored_index)

            loss_fct = CrossEntropyLoss(ignore_index=ignored_index)
            start_loss = loss_fct(start_logits, start_positions)
            end_loss = loss_fct(end_logits, end_positions)
            total_loss = (start_loss + end_loss) / 2

        if not return_dict:
            output = (start_logits, end_logits) + outputs[2:]
            return ((total_loss,) + output) if total_loss is not None else output

        return QuestionAnsweringModelOutput(
            loss=total_loss,
            start_logits=start_logits,
            end_logits=end_logits,
            hidden_states=outputs.hidden_states,
            attentions=outputs.attentions,
        )
